import logging

import pytest

import salt.netapi
from salt.exceptions import EauthAuthenticationError, SaltInvocationError
from tests.support.helpers import SaveRequestsPostHandler, Webserver
from tests.support.mock import patch

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.requires_sshd_server,
    pytest.mark.skipif(
        'grains["osfinger"].startswith(("Fedora Linux-40", "Ubuntu-24.04", "Arch Linux"))',
        reason="System ships with a version of python that is too recent for salt-ssh tests",
        # Actually, the problem is that the tornado we ship is not prepared for Python 3.12,
        # and it imports `ssl` and checks if the `match_hostname` function is defined, which
        # has been deprecated since Python 3.7, so, the logic goes into trying to import
        # backports.ssl-match-hostname which is not installed on the system.
    ),
    pytest.mark.timeout_unless_on_windows(120),
]

log = logging.getLogger(__name__)


@pytest.fixture
def client_config(client_config, known_hosts_file):
    client_config["known_hosts_file"] = str(known_hosts_file)
    client_config["netapi_enable_clients"] = ["ssh"]
    return client_config


@pytest.fixture
def client(client_config, salt_minion):
    return salt.netapi.NetapiClient(client_config)


@pytest.fixture
def rosters_dir(salt_ssh_roster_file):
    return str(salt_ssh_roster_file.parent)


@pytest.fixture
def ssh_priv_key(sshd_server):
    return str(sshd_server.config_dir / "client_key")


@pytest.fixture(scope="module")
def webserver():
    with Webserver(handler=SaveRequestsPostHandler) as server:
        yield server


@pytest.fixture(scope="module")
def webserver_root(webserver):
    return webserver.web_root


@pytest.fixture(scope="module")
def webserver_handler(webserver):
    return webserver.handler


@pytest.fixture(scope="module")
def salt_auth_account_1(salt_auth_account_1_factory):
    with salt_auth_account_1_factory as account:
        yield account


def test_ssh(client, auth_creds, salt_ssh_roster_file, rosters_dir, ssh_priv_key):
    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "roster_file": str(salt_ssh_roster_file),
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        **auth_creds,
    }

    ret = client.run(low)

    assert "localhost" in ret
    assert "return" in ret["localhost"]
    assert ret["localhost"]["return"] is True
    assert ret["localhost"]["id"] == "localhost"
    assert ret["localhost"]["fun"] == "test.ping"


def test_ssh_unauthenticated(client):
    low = {"client": "ssh", "tgt": "localhost", "fun": "test.ping"}

    with pytest.raises(EauthAuthenticationError):
        client.run(low)


def test_ssh_unauthenticated_raw_shell_curl(client, webserver_root, webserver_handler):

    fun = f"-o ProxyCommand curl {webserver_root}"
    low = {"client": "ssh", "tgt": "localhost", "fun": fun, "raw_shell": True}

    with pytest.raises(EauthAuthenticationError):
        client.run(low)

    assert webserver_handler.received_requests == []


def test_ssh_unauthenticated_raw_shell_touch(client, tmp_path):

    badfile = tmp_path / "badfile.txt"
    fun = f"-o ProxyCommand touch {badfile}"
    low = {"client": "ssh", "tgt": "localhost", "fun": fun, "raw_shell": True}

    with pytest.raises(EauthAuthenticationError):
        client.run(low)

    assert badfile.exists() is False


def test_ssh_authenticated_raw_shell_disabled(client, tmp_path):

    badfile = tmp_path / "badfile.txt"
    fun = f"-o ProxyCommand touch {badfile}"
    low = {"client": "ssh", "tgt": "localhost", "fun": fun, "raw_shell": True}

    with patch.dict(client.opts, {"netapi_allow_raw_shell": False}):
        with pytest.raises(EauthAuthenticationError):
            client.run(low)

    assert badfile.exists() is False


def test_ssh_disabled(client, auth_creds):
    low = {"client": "ssh", "tgt": "localhost", "fun": "test.ping", **auth_creds}

    ret = None
    with patch.dict(client.opts, {"netapi_enable_clients": []}):
        with pytest.raises(SaltInvocationError):
            ret = client.run(low)

    assert ret is None


@pytest.mark.timeout_unless_on_windows(360)
def test_shell_inject_ssh_priv(
    client, salt_ssh_roster_file, rosters_dir, tmp_path, salt_auto_account
):
    """
    Verify CVE-2020-16846 for ssh_priv variable
    """
    # ZDI-CAN-11143
    path = tmp_path / "test-11143"
    tgts = ["packages.broadcom.com", "www.zerodayinitiative.com"]
    ret = None
    for tgt in tgts:
        low = {
            "roster": "cache",
            "client": "ssh",
            "tgt": tgt,
            "ssh_priv": f"aaa|id>{path} #",
            "fun": "test.ping",
            "eauth": "auto",
            "username": salt_auto_account.username,
            "password": salt_auto_account.password,
            "roster_file": str(salt_ssh_roster_file),
            "rosters": [rosters_dir],
        }
        ret = client.run(low)
        if ret:
            break

    assert path.exists() is False
    assert ret
    assert not ret[tgt]["stdout"]
    assert ret[tgt]["stderr"]


def test_shell_inject_tgt(client, salt_ssh_roster_file, tmp_path, salt_auto_account):
    """
    Verify CVE-2020-16846 for tgt variable
    """
    # ZDI-CAN-11167
    path = tmp_path / "test-11167"
    low = {
        "roster": "cache",
        "client": "ssh",
        "tgt": f"root|id>{path} #@127.0.0.1",
        "roster_file": str(salt_ssh_roster_file),
        "rosters": "/",
        "fun": "test.ping",
        "eauth": "auto",
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "ignore_host_keys": True,
    }
    ret = client.run(low)
    assert path.exists() is False
    assert not ret["127.0.0.1"]["stdout"]
    assert ret["127.0.0.1"]["stderr"]


def test_shell_inject_ssh_options(
    client, salt_ssh_roster_file, tmp_path, salt_auto_account
):
    """
    Verify CVE-2020-16846 for ssh_options
    """
    # ZDI-CAN-11169
    path = tmp_path / "test-11169"
    low = {
        "roster": "cache",
        "client": "ssh",
        "tgt": "127.0.0.1",
        "renderer": "jinja|yaml",
        "fun": "test.ping",
        "eauth": "auto",
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "roster_file": str(salt_ssh_roster_file),
        "rosters": "/",
        "ssh_options": [f"|id>{path} #", "lol"],
    }
    ret = client.run(low)
    assert path.exists() is False
    assert not ret["127.0.0.1"]["stdout"]
    assert ret["127.0.0.1"]["stderr"]


def test_shell_inject_ssh_port(
    client, salt_ssh_roster_file, tmp_path, salt_auto_account
):
    """
    Verify CVE-2020-16846 for ssh_port variable
    """
    # ZDI-CAN-11172
    path = tmp_path / "test-11172"
    low = {
        "roster": "cache",
        "client": "ssh",
        "tgt": "127.0.0.1",
        "renderer": "jinja|yaml",
        "fun": "test.ping",
        "eauth": "auto",
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "roster_file": str(salt_ssh_roster_file),
        "rosters": "/",
        "ssh_port": f"hhhhh|id>{path} #",
        "ignore_host_keys": True,
    }
    ret = client.run(low)
    assert path.exists() is False
    assert not ret["127.0.0.1"]["stdout"]
    assert ret["127.0.0.1"]["stderr"]


def test_shell_inject_remote_port_forwards(
    client, salt_ssh_roster_file, tmp_path, salt_auto_account
):
    """
    Verify CVE-2020-16846 for remote_port_forwards variable
    """
    # ZDI-CAN-11173
    path = tmp_path / "test-1173"
    low = {
        "roster": "cache",
        "client": "ssh",
        "tgt": "127.0.0.1",
        "renderer": "jinja|yaml",
        "fun": "test.ping",
        "roster_file": str(salt_ssh_roster_file),
        "rosters": "/",
        "ssh_remote_port_forwards": f"hhhhh|id>{path} #, lol",
        "eauth": "auto",
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "ignore_host_keys": True,
    }
    ret = client.run(low)
    assert path.exists() is False
    assert not ret["127.0.0.1"]["stdout"]
    assert ret["127.0.0.1"]["stderr"]


def test_extra_mods(client, ssh_priv_key, rosters_dir, tmp_path, salt_auth_account_1):
    """
    validate input from extra_mods
    """
    path = tmp_path / "test_extra_mods"
    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "roster_file": "roster",
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        "eauth": "pam",
        "username": salt_auth_account_1.username,
        "password": salt_auth_account_1.password,
        "regen_thin": True,
        "thin_extra_mods": f"';touch {path};'",
    }

    ret = client.run(low)
    assert path.exists() is False
    assert "localhost" in ret
    assert "return" in ret["localhost"]
    assert ret["localhost"]["return"] is True


def test_ssh_auth_bypass(client, salt_ssh_roster_file):
    """
    CVE-2020-25592 - Bogus eauth raises exception.
    """
    low = {
        "roster": "cache",
        "client": "ssh",
        "tgt": "127.0.0.1",
        "renderer": "jinja|yaml",
        "fun": "test.ping",
        "roster_file": str(salt_ssh_roster_file),
        "rosters": "/",
        "eauth": "xx",
    }
    with pytest.raises(EauthAuthenticationError):
        client.run(low)


def test_ssh_auth_valid(client, ssh_priv_key, rosters_dir, salt_auth_account_1):
    """
    CVE-2020-25592 - Valid eauth works as expected.
    """
    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "roster_file": "roster",
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        "eauth": "pam",
        "username": salt_auth_account_1.username,
        "password": salt_auth_account_1.password,
    }
    ret = client.run(low)
    assert "localhost" in ret
    assert "return" in ret["localhost"]
    assert ret["localhost"]["return"] is True


def test_ssh_auth_invalid(client, rosters_dir, ssh_priv_key, salt_auth_account_1):
    """
    CVE-2020-25592 - Wrong password raises exception.
    """
    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "roster_file": "roster",
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        "eauth": "pam",
        "username": salt_auth_account_1.username,
        "password": "notvalidpassword",
    }
    with pytest.raises(EauthAuthenticationError):
        client.run(low)


def test_ssh_auth_invalid_acl(client, rosters_dir, ssh_priv_key, salt_auth_account_1):
    """
    CVE-2020-25592 - Eauth ACL enforced.
    """
    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "at.at",
        "args": ["12:05am", "echo foo"],
        "roster_file": "roster",
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        "eauth": "pam",
        "username": salt_auth_account_1.username,
        "password": "notvalidpassword",
    }
    with pytest.raises(EauthAuthenticationError):
        client.run(low)


def test_ssh_auth_token(client, rosters_dir, ssh_priv_key, salt_auth_account_1):
    """
    CVE-2020-25592 - Eauth tokens work as expected.
    """
    low = {
        "eauth": "pam",
        "username": salt_auth_account_1.username,
        "password": salt_auth_account_1.password,
    }
    ret = client.loadauth.mk_token(low)
    assert "token" in ret
    assert ret["token"]

    low = {
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "roster_file": "roster",
        "rosters": [rosters_dir],
        "ssh_priv": ssh_priv_key,
        "token": ret["token"],
    }
    ret = client.run(low)
    assert "localhost" in ret
    assert "return" in ret["localhost"]
    assert ret["localhost"]["return"] is True


def test_ssh_cve_2021_3197_a(
    client, rosters_dir, ssh_priv_key, salt_auth_account_1, tmp_path
):
    exploited_path = tmp_path / "exploited"
    assert exploited_path.exists() is False
    low = {
        "eauth": "auto",
        "username": salt_auth_account_1.username,
        "password": salt_auth_account_1.password,
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "ssh_port": f'22 -o ProxyCommand="touch {exploited_path}"',
        "ssh_priv": ssh_priv_key,
        "roster_file": "roster",
        "rosters": [rosters_dir],
    }
    ret = client.run(low)
    assert exploited_path.exists() is False
    assert "localhost" in ret
    assert ret["localhost"]["return"] is True


def test_ssh_cve_2021_3197_b(
    client, rosters_dir, ssh_priv_key, salt_auth_account_1, tmp_path
):
    exploited_path = tmp_path / "exploited"
    assert exploited_path.exists() is False
    low = {
        "eauth": "auto",
        "username": salt_auth_account_1.username,
        "password": salt_auth_account_1.password,
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "ssh_port": 22,
        "ssh_options": [f'ProxyCommand="touch {exploited_path}"'],
        "ssh_priv": ssh_priv_key,
        "roster_file": "roster",
        "rosters": [rosters_dir],
    }
    ret = client.run(low)
    assert exploited_path.exists() is False
    assert "localhost" in ret
    assert "return" in ret["localhost"]
    assert ret["localhost"]["return"] is True



"""
Proxy Minion interface module for managing Cisco Integrated Management Controller devices
=========================================================================================

.. versionadded:: 2018.3.0

:codeauthor: ``Spencer Ervin <spencer_ervin@hotmail.com>``
:maturity:   new
:depends:    none
:platform:   unix

This proxy minion enables Cisco Integrated Management Controller devices (hereafter referred to
as simply 'cimc' devices to be treated individually like a Salt Minion.

The cimc proxy leverages the XML API functionality on the Cisco Integrated Management Controller.
The Salt proxy must have access to the cimc on HTTPS (tcp/443).

More in-depth conceptual reading on Proxy Minions can be found in the
:ref:`Proxy Minion <proxy-minion>` section of Salt's
documentation.


Configuration
=============

To use this integration proxy module, please configure the following:

Pillar
------

Proxy minions get their configuration from Salt's Pillar. Every proxy must
have a stanza in Pillar and a reference in the Pillar top-file that matches
the ID.

.. code-block:: yaml

    proxy:
      proxytype: cimc
      host: <ip or dns name of cimc host>
      username: <cimc username>
      password: <cimc password>
      verify_ssl: True

proxytype
^^^^^^^^^

The ``proxytype`` key and value pair is critical, as it tells Salt which
interface to load from the ``proxy`` directory in Salt's install hierarchy,
or from ``/srv/salt/_proxy`` on the Salt Master (if you have created your
own proxy module, for example). To use this cimc Proxy Module, set this to
``cimc``.

host
^^^^

The location, or ip/dns, of the cimc host. Required.

username
^^^^^^^^

The username used to login to the cimc host. Required.

password
^^^^^^^^

The password used to login to the cimc host. Required.
"""

import logging
import re
import xml.etree.ElementTree as ET

import salt.exceptions

# This must be present or the Salt loader won't load this module.
__proxyenabled__ = ["cimc"]

# Variables are scoped to this module so we can have persistent data.
GRAINS_CACHE = {"vendor": "Cisco"}
DETAILS = {}

# Set up logging
log = logging.getLogger(__file__)

# Define the module's virtual name
__virtualname__ = "cimc"


def __virtual__():
    """
    Only return if all the modules are available.
    """
    return __virtualname__


def _validate_response_code(response_code_to_check, cookie_to_logout=None):
    formatted_response_code = str(response_code_to_check)
    if formatted_response_code not in ["200", "201", "202", "204"]:
        if cookie_to_logout:
            logout(cookie_to_logout)
        log.error("Received error HTTP status code: %s", formatted_response_code)
        raise salt.exceptions.CommandExecutionError(
            "Did not receive a valid response from host."
        )


def init(opts):
    """
    This function gets called when the proxy starts up.
    """
    log.debug("=== opts %s ===", opts)
    if "host" not in opts["proxy"]:
        log.critical("No 'host' key found in pillar for this proxy.")
        return False
    if "username" not in opts["proxy"]:
        log.critical("No 'username' key found in pillar for this proxy.")
        return False
    if "password" not in opts["proxy"]:
        log.critical("No 'passwords' key found in pillar for this proxy.")
        return False

    DETAILS["url"] = "https://{}/nuova".format(opts["proxy"]["host"])
    DETAILS["headers"] = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Content-Length": 62,
        "USER-Agent": "lwp-request/2.06",
    }

    # Set configuration details
    DETAILS["host"] = opts["proxy"]["host"]
    DETAILS["username"] = opts["proxy"].get("username")
    DETAILS["password"] = opts["proxy"].get("password")
    verify_ssl = opts["proxy"].get("verify_ssl")
    if verify_ssl is None:
        verify_ssl = True
    DETAILS["verify_ssl"] = verify_ssl

    # Ensure connectivity to the device
    log.debug("Attempting to connect to cimc proxy host.")
    get_config_resolver_class("computeRackUnit")
    log.debug("Successfully connected to cimc proxy host.")

    DETAILS["initialized"] = True


def set_config_modify(dn=None, inconfig=None, hierarchical=False):
    """
    The configConfMo method configures the specified managed object in a single subtree (for example, DN).
    """
    ret = {}
    cookie = logon()

    # Declare if the search contains hierarchical results.
    h = "false"
    if hierarchical is True:
        h = "true"

    payload = (
        '<configConfMo cookie="{}" inHierarchical="{}" dn="{}">'
        "<inConfig>{}</inConfig></configConfMo>".format(cookie, h, dn, inconfig)
    )
    r = __utils__["http.query"](
        DETAILS["url"],
        data=payload,
        method="POST",
        decode_type="plain",
        decode=True,
        verify_ssl=DETAILS["verify_ssl"],
        raise_error=True,
        status=True,
        headers=DETAILS["headers"],
    )

    _validate_response_code(r["status"], cookie)

    answer = re.findall(r"(<[\s\S.]*>)", r["text"])[0]
    items = ET.fromstring(answer)
    logout(cookie)
    for item in items:
        ret[item.tag] = prepare_return(item)
    return ret


def get_config_resolver_class(cid=None, hierarchical=False):
    """
    The configResolveClass method returns requested managed object in a given class.
    """
    ret = {}
    cookie = logon()

    # Declare if the search contains hierarchical results.
    h = "false"
    if hierarchical is True:
        h = "true"

    payload = (
        '<configResolveClass cookie="{}" inHierarchical="{}" classId="{}"/>'.format(
            cookie, h, cid
        )
    )
    r = __utils__["http.query"](
        DETAILS["url"],
        data=payload,
        method="POST",
        decode_type="plain",
        decode=True,
        verify_ssl=DETAILS["verify_ssl"],
        raise_error=True,
        status=True,
        headers=DETAILS["headers"],
    )

    _validate_response_code(r["status"], cookie)

    answer = re.findall(r"(<[\s\S.]*>)", r["text"])[0]
    items = ET.fromstring(answer)
    logout(cookie)

    for item in items:
        ret[item.tag] = prepare_return(item)
    return ret


def logon():
    """
    Logs into the cimc device and returns the session cookie.
    """
    content = {}
    payload = "<aaaLogin inName='{}' inPassword='{}'></aaaLogin>".format(
        DETAILS["username"], DETAILS["password"]
    )
    r = __utils__["http.query"](
        DETAILS["url"],
        data=payload,
        method="POST",
        decode_type="plain",
        decode=True,
        verify_ssl=DETAILS["verify_ssl"],
        raise_error=False,
        status=True,
        headers=DETAILS["headers"],
    )

    _validate_response_code(r["status"])

    answer = re.findall(r"(<[\s\S.]*>)", r["text"])[0]
    items = ET.fromstring(answer)
    for item in items.attrib:
        content[item] = items.attrib[item]

    if "outCookie" not in content:
        raise salt.exceptions.CommandExecutionError("Unable to log into proxy device.")

    return content["outCookie"]


def logout(cookie=None):
    """
    Closes the session with the device.
    """
    payload = '<aaaLogout cookie="{0}" inCookie="{0}"></aaaLogout>'.format(cookie)
    __utils__["http.query"](
        DETAILS["url"],
        data=payload,
        method="POST",
        decode_type="plain",
        decode=True,
        verify_ssl=DETAILS["verify_ssl"],
        raise_error=True,
        headers=DETAILS["headers"],
    )
    return


def prepare_return(x):
    """
    Converts the etree to dict
    """
    ret = {}
    for a in list(x):
        if a.tag not in ret:
            ret[a.tag] = []
        ret[a.tag].append(prepare_return(a))
    for a in x.attrib:
        ret[a] = x.attrib[a]
    return ret


def initialized():
    """
    Since grains are loaded in many different places and some of those
    places occur before the proxy can be initialized, return whether
    our init() function has been called
    """
    return DETAILS.get("initialized", False)


def grains():
    """
    Get the grains from the proxied device
    """
    if not DETAILS.get("grains_cache", {}):
        DETAILS["grains_cache"] = GRAINS_CACHE
        try:
            compute_rack = get_config_resolver_class("computeRackUnit", False)
            DETAILS["grains_cache"] = compute_rack["outConfigs"]["computeRackUnit"]
        except salt.exceptions.CommandExecutionError:
            pass
        except Exception as err:  # pylint: disable=broad-except
            log.error(err)
    return DETAILS["grains_cache"]


def grains_refresh():
    """
    Refresh the grains from the proxied device
    """
    DETAILS["grains_cache"] = None
    return grains()


def ping():
    """
    Returns true if the device is reachable, else false.
    """
    try:
        cookie = logon()
        logout(cookie)
    except salt.exceptions.CommandExecutionError:
        return False
    except Exception as err:  # pylint: disable=broad-except
        log.debug(err)
        return False
    return True


def shutdown():
    """
    Shutdown the connection to the proxy device. For this proxy,
    shutdown is a no-op.
    """
    log.debug("CIMC proxy shutdown() called.")



"""
Salt state to create, update VMware ESXi Virtual Machines.

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESX VSM module. Because the Salt extensions are newer and
    actively supported by VMware, they are more compatible with current versions
    of ESXi and they work well with the latest features in the VMware product
    line.


Dependencies
============

- pyVmomi
- jsonschema

States
======

vm_configured
-------------

Enforces correct virtual machine configuration. Creates, updates and registers
a virtual machine.

This state identifies the action which should be taken for the virtual machine
and applies that action via the create, update, register state functions.

Supported proxies: esxvm


Example:

1. Get the virtual machine ``my_vm`` status with an ``esxvm`` proxy:

Proxy minion configuration for ``esxvm`` proxy:

.. code-block:: yaml

    proxy:
      proxytype: esxvm
      datacenter: my_dc
      vcenter: vcenter.fake.com
      mechanism: sspi
      domain: fake.com
      principal: host

State configuration:

.. code-block:: yaml

      myvm_state:
        esxvm.vm_configured:
        - vm_name: my_vm
        - cpu: {{ {'count': 4, 'cores_per_socket': 2} }}
        - memory: {{ {'size': 16384, 'unit': 'MB'} }}
        - image: rhel7_64Guest
        - version: vmx-12
        - interfaces: {{ [{
                           'adapter': 'Network adapter 1',
                           'name': 'my_pg1',
                           'switch_type': 'distributed',
                           'adapter_type': 'vmxnet3',
                           'mac': '00:50:56:00:01:02,
                           'connectable': { 'start_connected': true,
                                            'allow_guest_control': true,
                                            'connected': true}},
                           {
                           'adapter': 'Network adapter 2',
                           'name': 'my_pg2',
                           'switch_type': 'distributed',
                           'adapter_type': 'vmxnet3',
                           'mac': '00:50:56:00:01:03',
                           'connectable': { 'start_connected': true,
                                            'allow_guest_control': true,
                                            'connected': true}}
                      ] }}
        - disks: {{ [{
                      'adapter': 'Hard disk 1',
                      'unit': 'MB',
                      'size': 51200,
                      'filename': 'my_vm/sda.vmdk',
                      'datastore': 'my_datastore',
                      'address': '0:0',
                      'thin_provision': true,
                      'eagerly_scrub': false,
                      'controller': 'SCSI controller 0'},
                      {
                      'adapter': 'Hard disk 2',
                      'unit': 'MB',
                      'size': 10240,
                      'filename': 'my_vm/sdb.vmdk',
                      'datastore': 'my_datastore',
                      'address': '0:1',
                      'thin_provision': true,
                      'eagerly_scrub': false,
                      'controller': 'SCSI controller 0'}
                ] }}
        - scsi_devices: {{ [{
                             'adapter': 'SCSI controller 0',
                             'type': 'paravirtual',
                             'bus_sharing': 'no_sharing',
                             'bus_number': 0}
                            ] }}
        - serial_ports: {{ [{
                             'adapter': 'Serial port 1',
                             'type': 'network',
                             'yield': false,
                             'backing': {
                                         'uri': 'my_uri',
                                         'direction': 'server',
                                         'filename': 'my_file'},
                              'connectable': {
                                              'start_connected': true,
                                              'allow_guest_control': true,
                                              'connected': true}}
                             ] }}
        - datacenter: {{ 'my_dc' }}
        - datastore: 'my_datastore'
        - placement: {{ {'cluster': 'my_cluster'} }}
        - cd_dvd_drives: {{ [] }}
        - advanced_configs: {{ {'my_param': '1'} }}
        - template: false
        - tools: false
        - power_on: false
        - deploy: false


vm_updated
----------

Updates a virtual machine to a given configuration.

vm_created
----------

Creates a virtual machine with a given configuration.

vm_registered
-------------

Registers a virtual machine with its configuration file path.

Dependencies
============

pyVmomi
-------

PyVmomi can be installed via pip:

.. code-block:: bash

    pip install pyVmomi

.. note::

    Version 6.0 of pyVmomi has some problems with SSL error handling on
    certain versions of Python. If using version 6.0 of pyVmomi, Python 2.6,
    Python 2.7.9, or newer must be present. This is due to an upstream
    dependency in pyVmomi 6.0 that is not supported in Python versions
    2.7 to 2.7.8. If the version of Python is not in the supported range,
    you will need to install an earlier version of pyVmomi.
    See `Issue #29537`_ for more information.

.. _Issue #29537: https://github.com/saltstack/salt/issues/29537

Based on the note above, to install an earlier version of pyVmomi than the
version currently listed in PyPi, run the following:

.. code-block:: bash

    pip install pyVmomi==6.0.0.2016.4

The 5.5.0.2014.1.1 is a known stable version that this original ESXi State
Module was developed against. To be able to connect through SSPI you must
use pyvmomi 6.0.0.2016.4 or above. The ESXVM State Module was tested with
this version.

About
-----

This state module was written to be used in conjunction with Salt's
:mod:`ESXi Proxy Minion <salt.proxy.esxi>` For a tutorial on how to use Salt's
ESXi Proxy Minion, please refer to the
:ref:`ESXi Proxy Minion Tutorial <tutorial-esxi-proxy>` for
configuration examples, dependency installation instructions, how to run remote
execution functions against ESXi hosts via a Salt Proxy Minion, and a larger state
example.
"""

import logging
import sys
from functools import wraps

import salt.exceptions
from salt.config.schemas.esxvm import ESXVirtualMachineConfigSchema

# External libraries
try:
    import jsonschema

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False

try:
    from pyVmomi import VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_JSONSCHEMA:
        return False, "State module did not load: jsonschema not found"
    if not HAS_PYVMOMI:
        return False, "State module did not load: pyVmomi not found"

    # We check the supported vim versions to infer the pyVmomi version
    if (
        "vim25/6.0" in VmomiSupport.versionMap
        and sys.version_info > (2, 7)
        and sys.version_info < (2, 7, 9)
    ):

        return (
            False,
            "State module did not load: Incompatible versions "
            "of Python and pyVmomi present. See Issue #29537.",
        )
    return True


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            3008,
            "The 'esxvm' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3008 in favor of the "
            "saltext.vmware Salt Extension. "
            "(https://github.com/saltstack/salt-ext-modules-vmware)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def vm_configured(
    name,
    vm_name,
    cpu,
    memory,
    image,
    version,
    interfaces,
    disks,
    scsi_devices,
    serial_ports,
    datacenter,
    datastore,
    placement,
    cd_dvd_drives=None,
    sata_controllers=None,
    advanced_configs=None,
    template=None,
    tools=True,
    power_on=False,
    deploy=False,
):
    """
    Selects the correct operation to be executed on a virtual machine, non
    existing machines will be created, existing ones will be updated if the
    config differs.
    """
    result = {"name": name, "result": None, "changes": {}, "comment": ""}

    log.trace("Validating virtual machine configuration")
    schema = ESXVirtualMachineConfigSchema.serialize()
    log.trace("schema = %s", schema)
    try:
        jsonschema.validate(
            {
                "vm_name": vm_name,
                "cpu": cpu,
                "memory": memory,
                "image": image,
                "version": version,
                "interfaces": interfaces,
                "disks": disks,
                "scsi_devices": scsi_devices,
                "serial_ports": serial_ports,
                "cd_dvd_drives": cd_dvd_drives,
                "sata_controllers": sata_controllers,
                "datacenter": datacenter,
                "datastore": datastore,
                "placement": placement,
                "template": template,
                "tools": tools,
                "power_on": power_on,
                "deploy": deploy,
            },
            schema,
        )
    except jsonschema.exceptions.ValidationError as exc:
        raise salt.exceptions.InvalidConfigError(exc)

    service_instance = __salt__["vsphere.get_service_instance_via_proxy"]()
    try:
        __salt__["vsphere.get_vm"](
            vm_name, vm_properties=["name"], service_instance=service_instance
        )
    except salt.exceptions.VMwareObjectRetrievalError:
        vm_file = __salt__["vsphere.get_vm_config_file"](
            vm_name, datacenter, placement, datastore, service_instance=service_instance
        )
        if vm_file:
            if __opts__["test"]:
                result.update(
                    {
                        "comment": "The virtual machine {} will be registered.".format(
                            vm_name
                        )
                    }
                )
                __salt__["vsphere.disconnect"](service_instance)
                return result
            result = vm_registered(
                vm_name, datacenter, placement, vm_file, power_on=power_on
            )
            return result
        else:
            if __opts__["test"]:
                result.update(
                    {
                        "comment": "The virtual machine {} will be created.".format(
                            vm_name
                        )
                    }
                )
                __salt__["vsphere.disconnect"](service_instance)
                return result
            if template:
                result = vm_cloned(name)
            else:
                result = vm_created(
                    name,
                    vm_name,
                    cpu,
                    memory,
                    image,
                    version,
                    interfaces,
                    disks,
                    scsi_devices,
                    serial_ports,
                    datacenter,
                    datastore,
                    placement,
                    cd_dvd_drives=cd_dvd_drives,
                    advanced_configs=advanced_configs,
                    power_on=power_on,
                )
            return result

    result = vm_updated(
        name,
        vm_name,
        cpu,
        memory,
        image,
        version,
        interfaces,
        disks,
        scsi_devices,
        serial_ports,
        datacenter,
        datastore,
        cd_dvd_drives=cd_dvd_drives,
        sata_controllers=sata_controllers,
        advanced_configs=advanced_configs,
        power_on=power_on,
    )
    __salt__["vsphere.disconnect"](service_instance)

    log.trace(result)
    return result


@_deprecation_message
def vm_cloned(name):
    """
    Clones a virtual machine from a template virtual machine if it doesn't
    exist and a template is defined.
    """
    result = {"name": name, "result": True, "changes": {}, "comment": ""}

    return result


@_deprecation_message
def vm_updated(
    name,
    vm_name,
    cpu,
    memory,
    image,
    version,
    interfaces,
    disks,
    scsi_devices,
    serial_ports,
    datacenter,
    datastore,
    cd_dvd_drives=None,
    sata_controllers=None,
    advanced_configs=None,
    power_on=False,
):
    """
    Updates a virtual machine configuration if there is a difference between
    the given and deployed configuration.
    """
    result = {"name": name, "result": None, "changes": {}, "comment": ""}

    service_instance = __salt__["vsphere.get_service_instance_via_proxy"]()
    current_config = __salt__["vsphere.get_vm_config"](
        vm_name, datacenter=datacenter, objects=False, service_instance=service_instance
    )

    diffs = __salt__["vsphere.compare_vm_configs"](
        {
            "name": vm_name,
            "cpu": cpu,
            "memory": memory,
            "image": image,
            "version": version,
            "interfaces": interfaces,
            "disks": disks,
            "scsi_devices": scsi_devices,
            "serial_ports": serial_ports,
            "datacenter": datacenter,
            "datastore": datastore,
            "cd_drives": cd_dvd_drives,
            "sata_controllers": sata_controllers,
            "advanced_configs": advanced_configs,
        },
        current_config,
    )
    if not diffs:
        result.update(
            {
                "result": True,
                "changes": {},
                "comment": f"Virtual machine {vm_name} is already up to date",
            }
        )
        return result

    if __opts__["test"]:
        comment = (
            "State vm_updated will update virtual machine '{}' "
            "in datacenter '{}':\n{}".format(
                vm_name,
                datacenter,
                "\n".join(
                    [
                        ":\n".join([key, difference.changes_str])
                        for key, difference in diffs.items()
                    ]
                ),
            )
        )
        result.update({"result": None, "comment": comment})
        __salt__["vsphere.disconnect"](service_instance)
        return result

    try:
        changes = __salt__["vsphere.update_vm"](
            vm_name,
            cpu,
            memory,
            image,
            version,
            interfaces,
            disks,
            scsi_devices,
            serial_ports,
            datacenter,
            datastore,
            cd_dvd_drives=cd_dvd_drives,
            sata_controllers=sata_controllers,
            advanced_configs=advanced_configs,
            service_instance=service_instance,
        )
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc)
        if service_instance:
            __salt__["vsphere.disconnect"](service_instance)
        result.update({"result": False, "comment": str(exc)})
        return result

    if power_on:
        try:
            __salt__["vsphere.power_on_vm"](vm_name, datacenter)
        except salt.exceptions.VMwarePowerOnError as exc:
            log.error("Error: %s", exc)
            if service_instance:
                __salt__["vsphere.disconnect"](service_instance)
            result.update({"result": False, "comment": str(exc)})
            return result
        changes.update({"power_on": True})

    __salt__["vsphere.disconnect"](service_instance)

    result = {
        "name": name,
        "result": True,
        "changes": changes,
        "comment": f"Virtual machine {vm_name} was updated successfully",
    }

    return result


@_deprecation_message
def vm_created(
    name,
    vm_name,
    cpu,
    memory,
    image,
    version,
    interfaces,
    disks,
    scsi_devices,
    serial_ports,
    datacenter,
    datastore,
    placement,
    ide_controllers=None,
    sata_controllers=None,
    cd_dvd_drives=None,
    advanced_configs=None,
    power_on=False,
):
    """
    Creates a virtual machine with the given properties if it doesn't exist.
    """
    result = {"name": name, "result": None, "changes": {}, "comment": ""}

    if __opts__["test"]:
        result["comment"] = f"Virtual machine {vm_name} will be created"
        return result

    service_instance = __salt__["vsphere.get_service_instance_via_proxy"]()
    try:
        info = __salt__["vsphere.create_vm"](
            vm_name,
            cpu,
            memory,
            image,
            version,
            datacenter,
            datastore,
            placement,
            interfaces,
            disks,
            scsi_devices,
            serial_ports=serial_ports,
            ide_controllers=ide_controllers,
            sata_controllers=sata_controllers,
            cd_drives=cd_dvd_drives,
            advanced_configs=advanced_configs,
            service_instance=service_instance,
        )
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc)
        if service_instance:
            __salt__["vsphere.disconnect"](service_instance)
        result.update({"result": False, "comment": str(exc)})
        return result

    if power_on:
        try:
            __salt__["vsphere.power_on_vm"](
                vm_name, datacenter, service_instance=service_instance
            )
        except salt.exceptions.VMwarePowerOnError as exc:
            log.error("Error: %s", exc)
            if service_instance:
                __salt__["vsphere.disconnect"](service_instance)
            result.update({"result": False, "comment": str(exc)})
            return result
        info["power_on"] = power_on

    changes = {"name": vm_name, "info": info}
    __salt__["vsphere.disconnect"](service_instance)
    result = {
        "name": name,
        "result": True,
        "changes": changes,
        "comment": f"Virtual machine {vm_name} created successfully",
    }

    return result


@_deprecation_message
def vm_registered(vm_name, datacenter, placement, vm_file, power_on=False):
    """
    Registers a virtual machine if the machine files are available on
    the main datastore.
    """
    result = {"name": vm_name, "result": None, "changes": {}, "comment": ""}

    vmx_path = f"{vm_file.folderPath}{vm_file.file[0].path}"
    log.trace("Registering virtual machine with vmx file: %s", vmx_path)
    service_instance = __salt__["vsphere.get_service_instance_via_proxy"]()
    try:
        __salt__["vsphere.register_vm"](
            vm_name, datacenter, placement, vmx_path, service_instance=service_instance
        )
    except salt.exceptions.VMwareMultipleObjectsError as exc:
        log.error("Error: %s", exc)
        if service_instance:
            __salt__["vsphere.disconnect"](service_instance)
        result.update({"result": False, "comment": str(exc)})
        return result
    except salt.exceptions.VMwareVmRegisterError as exc:
        log.error("Error: %s", exc)
        if service_instance:
            __salt__["vsphere.disconnect"](service_instance)
        result.update({"result": False, "comment": str(exc)})
        return result

    if power_on:
        try:
            __salt__["vsphere.power_on_vm"](
                vm_name, datacenter, service_instance=service_instance
            )
        except salt.exceptions.VMwarePowerOnError as exc:
            log.error("Error: %s", exc)
            if service_instance:
                __salt__["vsphere.disconnect"](service_instance)
            result.update({"result": False, "comment": str(exc)})
            return result
    __salt__["vsphere.disconnect"](service_instance)
    result.update(
        {
            "result": True,
            "changes": {"name": vm_name, "power_on": power_on},
            "comment": f"Virtual machine {vm_name} registered successfully",
        }
    )

    return result



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import glob
import os.path
import tempfile

import pytest

import salt.modules.qemu_nbd as qemu_nbd
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {qemu_nbd: {}}


def test_connect():
    """
    Test if it activate nbd for an image file.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(qemu_nbd.__salt__, {"cmd.run": mock}):
        with patch.object(os.path, "isfile", MagicMock(return_value=False)):
            assert qemu_nbd.connect("/tmp/image.raw") == ""
            assert qemu_nbd.connect("/tmp/image.raw") == ""

    with patch.object(os.path, "isfile", mock):
        with patch.object(glob, "glob", MagicMock(return_value=["/dev/nbd0"])):
            with patch.dict(
                qemu_nbd.__salt__,
                {"cmd.run": mock, "cmd.retcode": MagicMock(side_effect=[1, 0])},
            ):
                assert qemu_nbd.connect("/tmp/image.raw") == "/dev/nbd0"

            with patch.dict(
                qemu_nbd.__salt__,
                {"cmd.run": mock, "cmd.retcode": MagicMock(return_value=False)},
            ):
                assert qemu_nbd.connect("/tmp/image.raw") == ""


def test_mount():
    """
    Test if it pass in the nbd connection device location,
    mount all partitions and return a dict of mount points.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(qemu_nbd.__salt__, {"cmd.run": mock}):
        assert qemu_nbd.mount("/dev/nbd0") == {}


@pytest.mark.slow_test
def test_init():
    """
    Test if it mount the named image via qemu-nbd
    and return the mounted roots
    """
    mock = MagicMock(return_value=True)
    with patch.dict(qemu_nbd.__salt__, {"cmd.run": mock}):
        assert qemu_nbd.init("/srv/image.qcow2") == ""

    with patch.object(os.path, "isfile", mock), patch.object(
        glob, "glob", MagicMock(return_value=["/dev/nbd0"])
    ), patch.dict(
        qemu_nbd.__salt__,
        {
            "cmd.run": mock,
            "mount.mount": mock,
            "cmd.retcode": MagicMock(side_effect=[1, 0]),
        },
    ):
        expected = {
            os.sep.join([tempfile.gettempdir(), "nbd", "nbd0", "nbd0"]): "/dev/nbd0"
        }
        assert qemu_nbd.init("/srv/image.qcow2") == expected


def test_clear():
    """
    Test if it pass in the mnt dict returned from nbd_mount
    to unmount and disconnect the image from nbd.
    """
    mock_run = MagicMock(return_value=True)
    with patch.dict(
        qemu_nbd.__salt__,
        {"cmd.run": mock_run, "mount.umount": MagicMock(side_effect=[False, True])},
    ):
        assert qemu_nbd.clear({"/mnt/foo": "/dev/nbd0p1"}) == {
            "/mnt/foo": "/dev/nbd0p1"
        }
        assert qemu_nbd.clear({"/mnt/foo": "/dev/nbd0p1"}) == {}



import logging

import pytest
import requests.exceptions

# pylint: disable=unused-import
from tests.support.pytest.vault import (
    vault_container_version,
    vault_delete_secret,
    vault_environ,
    vault_list_secrets,
    vault_read_secret,
    vault_write_secret,
)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        }
    }


@pytest.fixture
def vault(loaders):
    return loaders.utils.vault


@pytest.fixture(scope="module", autouse=True)
def vault_testing_data(vault_container_version):
    vault_write_secret("secret/utils/read", success="yup")
    vault_write_secret("secret/utils/deleteme", success="nope")
    try:
        yield
    finally:
        secret_path = "secret/utils"
        for secret in vault_list_secrets(secret_path):
            vault_delete_secret(f"{secret_path}/{secret}", metadata=True)


def test_make_request_get_unauthd(vault):
    """
    Test that unauthenticated GET requests are possible
    """
    res = vault.make_request("GET", "/v1/sys/health")
    assert res.status_code == 200
    assert res.json()
    assert "initialized" in res.json()


def test_make_request_get_authd(vault, vault_container_version):
    """
    Test that authenticated GET requests are possible
    """
    endpoint = "secret/utils/read"
    if vault_container_version in ["1.3.1", "latest"]:
        endpoint = "secret/data/utils/read"

    res = vault.make_request("GET", f"/v1/{endpoint}")
    assert res.status_code == 200
    data = res.json()["data"]
    if vault_container_version in ["1.3.1", "latest"]:
        data = data["data"]
    assert "success" in data
    assert data["success"] == "yup"


def test_make_request_post_json(vault, vault_container_version):
    """
    Test that POST requests are possible with json param
    """
    data = {"success": "yup"}
    endpoint = "secret/utils/write"

    if vault_container_version in ["1.3.1", "latest"]:
        data = {"data": data}
        endpoint = "secret/data/utils/write"
    res = vault.make_request("POST", f"/v1/{endpoint}", json=data)
    assert res.status_code in [200, 204]
    assert vault_read_secret("secret/utils/write") == {"success": "yup"}


def test_make_request_post_data(vault, vault_container_version):
    """
    Test that POST requests are possible with data param
    """
    data = '{"success": "yup_data"}'
    endpoint = "secret/utils/write"

    if vault_container_version in ["1.3.1", "latest"]:
        data = '{"data": {"success": "yup_data"}}'
        endpoint = "secret/data/utils/write"
    res = vault.make_request("POST", f"/v1/{endpoint}", data=data)
    assert res.status_code in [200, 204]
    assert vault_read_secret("secret/utils/write") == {"success": "yup_data"}


def test_make_request_delete(vault, vault_container_version):
    """
    Test that DELETE requests are possible
    """
    endpoint = "secret/utils/deleteme"
    if vault_container_version in ["1.3.1", "latest"]:
        endpoint = "secret/data/utils/deleteme"

    res = vault.make_request("DELETE", f"/v1/{endpoint}")
    assert res.status_code in [200, 204]
    assert vault_read_secret("secret/utils/deleteme") is None


def test_make_request_list(vault, vault_container_version):
    """
    Test that LIST requests are possible
    """
    endpoint = "secret/utils"
    if vault_container_version in ["1.3.1", "latest"]:
        endpoint = "secret/metadata/utils"

    res = vault.make_request("LIST", f"/v1/{endpoint}")
    assert res.status_code == 200
    assert res.json()["data"]["keys"] == vault_list_secrets("secret/utils")


def test_make_request_token_override(vault, vault_container_version):
    """
    Test that overriding the token in use is possible
    """
    endpoint = "secret/utils/read"
    if vault_container_version in ["1.3.1", "latest"]:
        endpoint = "secret/data/utils/read"

    res = vault.make_request("GET", f"/v1/{endpoint}", token="invalid")
    assert res.status_code == 403


def test_make_request_url_override(vault, vault_container_version):
    """
    Test that overriding the server URL is possible
    """
    endpoint = "secret/utils/read"
    if vault_container_version in ["1.3.1", "latest"]:
        endpoint = "secret/data/utils/read"

    with pytest.raises(
        requests.exceptions.ConnectionError, match=".*Max retries exceeded with url:.*"
    ):
        vault.make_request(
            "GET", f"/v1/{endpoint}", vault_url="http://127.0.0.1:1", timeout=2
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    TestCase for salt.modules.smtp
"""

import pytest

import salt.modules.smtp as smtp
from tests.support.mock import MagicMock, patch


class SMTPRecipientsRefused(Exception):
    """
    Mock SMTPRecipientsRefused class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class SMTPHeloError(Exception):
    """
    Mock SMTPHeloError class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class SMTPSenderRefused(Exception):
    """
    Mock SMTPSenderRefused class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class SMTPDataError(Exception):
    """
    Mock SMTPDataError class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class SMTPException(Exception):
    """
    Mock SMTPException class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class SMTPAuthenticationError(Exception):
    """
    Mock SMTPAuthenticationError class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class MockSMTPSSL:
    """
    Mock SMTP_SSL class
    """

    flag = None

    def __init__(self, server):
        pass

    def sendmail(self, sender, recipient, msg):
        """
        Mock sendmail method
        """
        if self.flag == 1:
            raise SMTPRecipientsRefused("All recipients were refused.")
        elif self.flag == 2:
            raise SMTPHeloError("Helo error")
        elif self.flag == 3:
            raise SMTPSenderRefused("Sender Refused")
        elif self.flag == 4:
            raise SMTPDataError("Data error")
        return (sender, recipient, msg)

    def login(self, username, password):
        """
        Mock login method
        """
        if self.flag == 5:
            raise SMTPAuthenticationError("SMTP Authentication Failure")
        return (username, password)

    @staticmethod
    def quit():
        """
        Mock quit method
        """
        return True


class MockSMTP:
    """
    Mock SMTP class
    """

    flag = None

    def __init__(self, server):
        pass

    @staticmethod
    def ehlo():
        """
        Mock ehlo method
        """
        return True

    @staticmethod
    def has_extn(name):
        """
        Mock has_extn method
        """
        return name

    def starttls(self):
        """
        Mock starttls method
        """
        if self.flag == 1:
            raise SMTPHeloError("Helo error")
        elif self.flag == 2:
            raise SMTPException("Exception error")
        elif self.flag == 3:
            raise RuntimeError
        return True

    def sendmail(self, sender, recipient, msg):
        """
        Mock sendmail method
        """
        if self.flag == 1:
            raise SMTPRecipientsRefused("All recipients were refused.")
        elif self.flag == 2:
            raise SMTPHeloError("Helo error")
        elif self.flag == 3:
            raise SMTPSenderRefused("Sender Refused")
        elif self.flag == 4:
            raise SMTPDataError("Data error")
        return (sender, recipient, msg)

    @staticmethod
    def quit():
        """
        Mock quit method
        """
        return True


class MockGaierror(Exception):
    """
    Mock MockGaierror class
    """

    def __init__(self, msg):
        super().__init__(msg)
        self.smtp_error = msg


class MockSocket:
    """
    Mock Socket class
    """

    def __init__(self):
        self.gaierror = MockGaierror


class MockSmtplib:
    """
    Mock smtplib class
    """

    flag = None

    def __init__(self):
        self.SMTPRecipientsRefused = SMTPRecipientsRefused
        self.SMTPHeloError = SMTPHeloError
        self.SMTPSenderRefused = SMTPSenderRefused
        self.SMTPDataError = SMTPDataError
        self.SMTPException = SMTPException
        self.SMTPAuthenticationError = SMTPAuthenticationError
        self.server = None

    def SMTP_SSL(self, server):
        """
        Mock SMTP_SSL method
        """
        self.server = server
        if self.flag == 1:
            raise MockGaierror("gaierror")
        return MockSMTPSSL("server")

    def SMTP(self, server):
        """
        Mock SMTP method
        """
        self.server = server
        if self.flag == 1:
            raise MockGaierror("gaierror")
        return MockSMTP("server")


@pytest.fixture
def configure_loader_modules():
    return {smtp: {"socket": MockSocket(), "smtplib": MockSmtplib()}}


# 'send_msg' function tests: 1


def test_send_msg():
    """
    Tests if it send a message to an SMTP recipient.
    """
    mock = MagicMock(
        return_value={
            "smtp.server": "",
            "smtp.tls": "True",
            "smtp.sender": "",
            "smtp.username": "",
            "smtp.password": "",
        }
    )
    with patch.dict(smtp.__salt__, {"config.option": mock}):
        assert smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTPSSL.flag = 1
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTPSSL.flag = 2
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTPSSL.flag = 3
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTPSSL.flag = 4
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

    mock = MagicMock(
        return_value={
            "smtp.server": "",
            "smtp.tls": "",
            "smtp.sender": "",
            "smtp.username": "",
            "smtp.password": "",
        }
    )
    with patch.dict(smtp.__salt__, {"config.option": mock}):
        MockSMTPSSL.flag = 5
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            username="myuser",
            password="verybadpass",
            sender="admin@example.com",
            server="smtp.domain.com",
        )

        MockSMTP.flag = 1
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTP.flag = 2
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSMTP.flag = 3
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )

        MockSmtplib.flag = 1
        assert not smtp.send_msg(
            "admin@example.com",
            "This is a salt module test",
            profile="my-smtp-account",
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.states.keystone
"""

import pytest

import salt.states.keystone as keystone
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {keystone: {}}


def test_user_present():
    """
    Test to ensure that the keystone user is present
    with the specified properties.
    """
    name = "nova"
    password = "$up3rn0v4"
    email = "nova@domain.com"
    tenant = "demo"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock_f = MagicMock(return_value=False)
    mock_lst = MagicMock(return_value=["Error"])
    with patch.dict(keystone.__salt__, {"keystone.tenant_get": mock_lst}):
        comt = f'Tenant / project "{tenant}" does not exist'
        ret.update({"comment": comt})
        assert keystone.user_present(name, password, email, tenant) == ret

    mock_dict = MagicMock(
        side_effect=[
            {name: {"email": "a@a.com"}},
            {name: {"email": email, "enabled": False}},
            {name: {"email": email, "enabled": True}},
            {name: {"email": email, "enabled": True}},
            {"Error": "error"},
            {"Error": "error"},
        ]
    )
    mock_l = MagicMock(return_value={tenant: {"id": "abc"}})
    with patch.dict(
        keystone.__salt__,
        {
            "keystone.user_get": mock_dict,
            "keystone.tenant_get": mock_l,
            "keystone.user_verify_password": mock_f,
            "keystone.user_create": mock_f,
        },
    ):
        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'User "{name}" will be updated'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {
                        "Email": "Will be updated",
                        "Enabled": "Will be True",
                        "Password": "Will be updated",
                    },
                }
            )
            assert keystone.user_present(name, password, email) == ret

            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {
                        "Enabled": "Will be True",
                        "Password": "Will be updated",
                    },
                }
            )
            assert keystone.user_present(name, password, email) == ret

            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {
                        "Tenant": 'Will be added to "demo" tenant',
                        "Password": "Will be updated",
                    },
                }
            )
            assert keystone.user_present(name, password, email, tenant) == ret

            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"Password": "Will be updated"},
                }
            )
            assert keystone.user_present(name, password, email) == ret

            comt = 'Keystone user "nova" will be added'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"User": "Will be created"},
                }
            )
            assert keystone.user_present(name, password, email) == ret

        with patch.dict(keystone.__opts__, {"test": False}):
            comt = f"Keystone user {name} has been added"
            ret.update(
                {"comment": comt, "result": True, "changes": {"User": "Created"}}
            )
            assert keystone.user_present(name, password, email) == ret


def test_user_absent():
    """
    Test to ensure that the keystone user is absent.
    """
    name = "nova"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'User "{name}" is already absent',
    }

    mock_lst = MagicMock(side_effect=[["Error"], []])
    with patch.dict(keystone.__salt__, {"keystone.user_get": mock_lst}):
        assert keystone.user_absent(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'User "{name}" will be deleted'
            ret.update({"comment": comt, "result": None})
            assert keystone.user_absent(name) == ret


def test_tenant_present():
    """
    Test to ensures that the keystone tenant exists
    """
    name = "nova"
    description = "OpenStack Compute Service"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Tenant / project "{name}" already exists',
    }

    mock_dict = MagicMock(
        side_effect=[
            {name: {"description": "desc"}},
            {name: {"description": description, "enabled": False}},
            {"Error": "error"},
            {"Error": "error"},
        ]
    )
    mock_t = MagicMock(return_value=True)
    with patch.dict(
        keystone.__salt__,
        {"keystone.tenant_get": mock_dict, "keystone.tenant_create": mock_t},
    ):
        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Tenant / project "{name}" will be updated'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"Description": "Will be updated"},
                }
            )
            assert keystone.tenant_present(name) == ret

            comt = f'Tenant / project "{name}" will be updated'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"Enabled": "Will be True"},
                }
            )
            assert keystone.tenant_present(name, description) == ret

            comt = f'Tenant / project "{name}" will be added'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"Tenant": "Will be created"},
                }
            )
            assert keystone.tenant_present(name) == ret

        with patch.dict(keystone.__opts__, {"test": False}):
            comt = f'Tenant / project "{name}" has been added'
            ret.update(
                {"comment": comt, "result": True, "changes": {"Tenant": "Created"}}
            )
            assert keystone.tenant_present(name) == ret


def test_tenant_absent():
    """
    Test to ensure that the keystone tenant is absent.
    """
    name = "nova"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Tenant / project "{name}" is already absent',
    }

    mock_lst = MagicMock(side_effect=[["Error"], []])
    with patch.dict(keystone.__salt__, {"keystone.tenant_get": mock_lst}):
        assert keystone.tenant_absent(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Tenant / project "{name}" will be deleted'
            ret.update({"comment": comt, "result": None})
            assert keystone.tenant_absent(name) == ret


def test_role_present():
    """
    Test to ensures that the keystone role exists
    """
    name = "nova"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Role "{name}" already exists',
    }

    mock_lst = MagicMock(side_effect=[[], ["Error"]])
    with patch.dict(keystone.__salt__, {"keystone.role_get": mock_lst}):
        assert keystone.role_present(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Role "{name}" will be added'
            ret.update({"comment": comt, "result": None})
            assert keystone.role_present(name) == ret


def test_role_absent():
    """
    Test to ensure that the keystone role is absent.
    """
    name = "nova"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Role "{name}" is already absent',
    }

    mock_lst = MagicMock(side_effect=[["Error"], []])
    with patch.dict(keystone.__salt__, {"keystone.role_get": mock_lst}):
        assert keystone.role_absent(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Role "{name}" will be deleted'
            ret.update({"comment": comt, "result": None})
            assert keystone.role_absent(name) == ret


def test_service_present():
    """
    Test to ensure service present in Keystone catalog
    """
    name = "nova"
    service_type = "compute"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Service "{name}" already exists',
    }

    mock_lst = MagicMock(side_effect=[[], ["Error"]])
    with patch.dict(keystone.__salt__, {"keystone.service_get": mock_lst}):
        assert keystone.service_present(name, service_type) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Service "{name}" will be added'
            ret.update({"comment": comt, "result": None})
            assert keystone.service_present(name, service_type) == ret


def test_service_absent():
    """
    Test to ensure that the service doesn't exist in Keystone catalog
    """
    name = "nova"

    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f'Service "{name}" is already absent',
    }

    mock_lst = MagicMock(side_effect=[["Error"], []])
    with patch.dict(keystone.__salt__, {"keystone.service_get": mock_lst}):
        assert keystone.service_absent(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Service "{name}" will be deleted'
            ret.update({"comment": comt, "result": None})
            assert keystone.service_absent(name) == ret


def test_endpoint_present():
    """
    Test to ensure the specified endpoints exists for service
    """
    name = "nova"
    region = "RegionOne"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    endpoint = {
        "adminurl": None,
        "region": None,
        "internalurl": None,
        "publicurl": None,
        "id": 1,
        "service_id": None,
    }

    mock_lst = MagicMock(
        side_effect=[endpoint, ["Error"], {"id": 1, "service_id": None}, []]
    )
    mock = MagicMock(return_value=True)
    with patch.dict(
        keystone.__salt__,
        {"keystone.endpoint_get": mock_lst, "keystone.endpoint_create": mock},
    ):

        comt = f'Endpoint for service "{name}" already exists'
        ret.update({"comment": comt, "result": True, "changes": {}})
        assert keystone.endpoint_present(name) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Endpoint for service "{name}" will be added'
            ret.update(
                {
                    "comment": comt,
                    "result": None,
                    "changes": {"Endpoint": "Will be created"},
                }
            )
            assert keystone.endpoint_present(name) == ret

            comt = f'Endpoint for service "{name}" already exists'
            ret.update({"comment": comt, "result": True, "changes": {}})
            assert keystone.endpoint_present(name) == ret

        with patch.dict(keystone.__opts__, {"test": False}):
            comt = f'Endpoint for service "{name}" has been added'
            ret.update({"comment": comt, "result": True, "changes": True})
            assert keystone.endpoint_present(name) == ret


def test_endpoint_absent():
    """
    Test to ensure that the endpoint for a service doesn't
     exist in Keystone catalog
    """
    name = "nova"
    region = "RegionOne"
    comment = f'Endpoint for service "{name}" is already absent'
    ret = {"name": name, "changes": {}, "result": True, "comment": comment}

    mock_lst = MagicMock(side_effect=[[], ["Error"]])
    with patch.dict(keystone.__salt__, {"keystone.endpoint_get": mock_lst}):
        assert keystone.endpoint_absent(name, region) == ret

        with patch.dict(keystone.__opts__, {"test": True}):
            comt = f'Endpoint for service "{name}" will be deleted'
            ret.update({"comment": comt, "result": None})
            assert keystone.endpoint_absent(name, region) == ret



"""
Use `Varstack <https://github.com/conversis/varstack>`_ data as a Pillar source

Configuring Varstack
====================

Using varstack in Salt is fairly simple. Just put the following into the
config file of your master:

.. code-block:: yaml

    ext_pillar:
      - varstack: /etc/varstack.yaml

Varstack will then use /etc/varstack.yaml to determine which configuration
data to return as pillar information. From there you can take a look at the
`README <https://github.com/conversis/varstack/blob/master/README.md>`_ of
varstack on how this file is evaluated.
"""

try:
    import varstack
except ImportError:
    varstack = None

# Define the module's virtual name
__virtualname__ = "varstack"


def __virtual__():
    return (
        varstack and __virtualname__ or False,
        "The varstack module could not be loaded: varstack dependency is missing.",
    )


def ext_pillar(
    minion_id, pillar, conf  # pylint: disable=W0613  # pylint: disable=W0613
):
    """
    Parse varstack data and return the result
    """
    vs = varstack.Varstack(config_filename=conf)
    return vs.evaluate(__grains__)



import logging

import pytest

import salt.config
import salt.state

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.core_test,
]


@pytest.fixture
def minion_config(minion_opts):
    minion_opts["file_client"] = "local"
    minion_opts["id"] = "foo01"
    return minion_opts


def test_global_state_conditions_unconfigured(minion_config):
    state_obj = salt.state.State(minion_config)
    ret = state_obj._match_global_state_conditions(
        "test.succeed_with_changes", "test", "mytest"
    )
    assert ret is None


@pytest.mark.parametrize("condition", [["foo01"], "foo01"])
def test_global_state_conditions_match(minion_config, condition):
    minion_config["global_state_conditions"] = {
        "test": condition,
    }
    state_obj = salt.state.State(minion_config)
    ret = state_obj._match_global_state_conditions(
        "test.succeed_with_changes", "test", "mytest"
    )
    assert ret is None


def test_global_state_conditions_no_match(minion_config):
    minion_config["global_state_conditions"] = {
        "test.succeed_with_changes": ["bar01"],
    }
    state_obj = salt.state.State(minion_config)
    ret = state_obj._match_global_state_conditions(
        "test.succeed_with_changes", "test", "mytest"
    )
    assert ret == {
        "changes": {},
        "comment": "Failed to meet global state conditions. State not called.",
        "name": "mytest",
        "result": None,
    }


def test_global_state_conditions_match_one_of_many(minion_config):
    minion_config["global_state_conditions"] = {
        "test.succeed_with_changes": ["bar01"],
        "test": ["baz01"],
        "*": ["foo01"],
    }
    state_obj = salt.state.State(minion_config)
    ret = state_obj._match_global_state_conditions(
        "test.succeed_with_changes", "test", "mytest"
    )
    assert ret is None



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases to salt.modules.sysbench
"""

import pytest

import salt.modules.sysbench as sysbench
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {sysbench: {}}


def test_cpu():
    """
    Test to tests to the CPU performance of minions.
    """
    with patch.dict(sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}):
        with patch.object(sysbench, "_parser", return_value={"A": "a"}):
            assert sysbench.cpu() == {
                "Prime numbers limit: 500": {"A": "a"},
                "Prime numbers limit: 5000": {"A": "a"},
                "Prime numbers limit: 2500": {"A": "a"},
                "Prime numbers limit: 1000": {"A": "a"},
            }


def test_threads():
    """
    Test to this tests the performance of the processor's scheduler
    """
    with patch.dict(sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}):
        with patch.object(sysbench, "_parser", return_value={"A": "a"}):
            assert sysbench.threads() == {
                "Yields: 500 Locks: 8": {"A": "a"},
                "Yields: 200 Locks: 4": {"A": "a"},
                "Yields: 1000 Locks: 16": {"A": "a"},
                "Yields: 100 Locks: 2": {"A": "a"},
            }


def test_mutex():
    """
    Test to tests the implementation of mutex
    """
    with patch.dict(sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}):
        with patch.object(sysbench, "_parser", return_value={"A": "a"}):
            assert sysbench.mutex() == {
                "Mutex: 1000 Locks: 25000 Loops: 10000": {"A": "a"},
                "Mutex: 50 Locks: 10000 Loops: 2500": {"A": "a"},
                "Mutex: 1000 Locks: 10000 Loops: 5000": {"A": "a"},
                "Mutex: 500 Locks: 50000 Loops: 5000": {"A": "a"},
                "Mutex: 500 Locks: 25000 Loops: 2500": {"A": "a"},
                "Mutex: 500 Locks: 10000 Loops: 10000": {"A": "a"},
                "Mutex: 50 Locks: 50000 Loops: 10000": {"A": "a"},
                "Mutex: 1000 Locks: 50000 Loops: 2500": {"A": "a"},
                "Mutex: 50 Locks: 25000 Loops: 5000": {"A": "a"},
            }


def test_memory():
    """
    Test to this tests the memory for read and write operations.
    """
    with patch.dict(sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}):
        with patch.object(sysbench, "_parser", return_value={"A": "a"}):
            assert sysbench.memory() == {
                "Operation: read Scope: local": {"A": "a"},
                "Operation: write Scope: local": {"A": "a"},
                "Operation: read Scope: global": {"A": "a"},
                "Operation: write Scope: global": {"A": "a"},
            }


def test_fileio():
    """
    Test to this tests for the file read and write operations
    """
    with patch.dict(sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}):
        with patch.object(sysbench, "_parser", return_value={"A": "a"}):
            assert sysbench.fileio() == {
                "Mode: seqrd": {"A": "a"},
                "Mode: seqwr": {"A": "a"},
                "Mode: rndrd": {"A": "a"},
                "Mode: rndwr": {"A": "a"},
                "Mode: seqrewr": {"A": "a"},
                "Mode: rndrw": {"A": "a"},
            }


def test_ping():
    """
    Test to ping
    """
    assert sysbench.ping()



import pytest

import salt.utils.platform

if salt.utils.platform.is_windows():
    from salt.platform import win

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.mark.parametrize(
    "command, expected",
    [
        ("whoami", "whoami"),
        ("hostname", "hostname"),
        ("cmd /c hostname", "cmd /c hostname"),
        ("echo foo", 'cmd /c "echo foo"'),
        ('cmd /c "echo foo"', 'cmd /c "echo foo"'),
        ("whoami && echo foo", 'cmd /c "whoami && echo foo"'),
        ("whoami || echo foo", 'cmd /c "whoami || echo foo"'),
    ],
)
def test_prepend_cmd(command, expected):
    """
    Test that the command is prepended with "cmd /c" where appropriate
    """
    result = win.prepend_cmd(command)
    assert result == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.alternatives as alternatives
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {alternatives: {}}


# 'install' function tests: 1


def test_install():
    """
    Test to install new alternative for defined <name>
    """
    name = "pager"
    link = "/usr/bin/pager"
    path = "/usr/bin/less"
    priority = 5

    ret = {
        "name": name,
        "link": link,
        "path": path,
        "priority": priority,
        "result": None,
        "changes": {},
        "comment": "",
    }

    bad_link = "/bin/pager"
    err = f"the primary link for {name} must be {link}"

    mock_cinst = MagicMock(side_effect=[True, False])
    mock_cexist = MagicMock(
        side_effect=[True, False, False, True, False, False, False, True]
    )
    mock_out = MagicMock(side_effect=["", err, ""])
    mock_path = MagicMock(return_value=path)
    mock_link = MagicMock(return_value=link)
    with patch.dict(
        alternatives.__salt__,
        {
            "alternatives.check_installed": mock_cinst,
            "alternatives.check_exists": mock_cexist,
            "alternatives.install": mock_out,
            "alternatives.show_current": mock_path,
            "alternatives.show_link": mock_link,
        },
    ):
        comt = f"Alternative {path} for {name} is already registered"
        ret.update({"comment": comt, "result": True})
        assert alternatives.install(name, link, path, priority) == ret

        comt = "Alternative will be set for {} to {} with priority {}".format(
            name, path, priority
        )
        ret.update({"comment": comt, "result": None})
        with patch.dict(alternatives.__opts__, {"test": True}):
            assert alternatives.install(name, link, path, priority) == ret

        comt = "Alternative for {} set to path {} with priority {}".format(
            name, path, priority
        )
        ret.update(
            {
                "comment": comt,
                "result": True,
                "changes": {
                    "name": name,
                    "link": link,
                    "path": path,
                    "priority": priority,
                },
            }
        )
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.install(name, link, path, priority) == ret

        comt = f"Alternative for {name} not installed: {err}"
        ret.update({"comment": comt, "result": False, "changes": {}, "link": bad_link})
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.install(name, bad_link, path, priority) == ret

        comt = "Alternative {} for {} registered with priority {} and not set to default".format(
            path, name, priority
        )
        ret.update(
            {
                "comment": comt,
                "result": True,
                "changes": {
                    "name": name,
                    "link": link,
                    "path": path,
                    "priority": priority,
                },
                "link": link,
            }
        )
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.install(name, link, path, priority) == ret


# 'remove' function tests: 1


def test_remove():
    """
    Test to removes installed alternative for defined <name> and <path>
    or fallback to default alternative, if some defined before.
    """
    name = "pager"
    path = "/usr/bin/less"

    ret = {"name": name, "path": path, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, True, True, False, False])
    mock_bool = MagicMock(return_value=True)
    mock_show = MagicMock(side_effect=[False, True, True, False])
    with patch.dict(
        alternatives.__salt__,
        {
            "alternatives.check_exists": mock,
            "alternatives.show_current": mock_show,
            "alternatives.remove": mock_bool,
        },
    ):
        comt = f"Alternative for {name} will be removed"
        ret.update({"comment": comt})
        with patch.dict(alternatives.__opts__, {"test": True}):
            assert alternatives.remove(name, path) == ret

        comt = f"Alternative for {name} removed"
        ret.update({"comment": comt, "result": True})
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.remove(name, path) == ret

        comt = "Alternative for pager removed. Falling back to path True"
        ret.update({"comment": comt, "result": True, "changes": {"path": True}})
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.remove(name, path) == ret

        comt = f"Alternative for {name} is set to it's default path True"
        ret.update({"comment": comt, "result": True, "changes": {}})
        assert alternatives.remove(name, path) == ret

        comt = f"Alternative for {name} doesn't exist"
        ret.update({"comment": comt, "result": False})
        assert alternatives.remove(name, path) == ret


# 'auto' function tests: 1


def test_auto():
    """
    Test to instruct alternatives to use the highest priority
    path for <name>
    """
    name = "pager"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[" auto mode", " ", " "])
    mock_auto = MagicMock(return_value=True)
    with patch.dict(
        alternatives.__salt__,
        {"alternatives.display": mock, "alternatives.auto": mock_auto},
    ):
        comt = f"{name} already in auto mode"
        ret.update({"comment": comt})
        assert alternatives.auto(name) == ret

        comt = f"{name} will be put in auto mode"
        ret.update({"comment": comt, "result": None})
        with patch.dict(alternatives.__opts__, {"test": True}):
            assert alternatives.auto(name) == ret

        ret.update({"comment": "", "result": True, "changes": {"result": True}})
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.auto(name) == ret


# 'set_' function tests: 1


def test_set():
    """
    Test to sets alternative for <name> to <path>, if <path> is defined
    as an alternative for <name>.
    """
    name = "pager"
    path = "/usr/bin/less"

    ret = {"name": name, "path": path, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[path, path, ""])
    mock_bool = MagicMock(return_value=True)
    mock_show = MagicMock(side_effect=[path, False, False, False, False])
    with patch.dict(
        alternatives.__salt__,
        {
            "alternatives.display": mock,
            "alternatives.show_current": mock_show,
            "alternatives.set": mock_bool,
        },
    ):
        comt = f"Alternative for {name} already set to {path}"
        ret.update({"comment": comt})
        assert alternatives.set_(name, path) == ret

        comt = f"Alternative for {name} will be set to path /usr/bin/less"
        ret.update({"comment": comt, "result": None})
        with patch.dict(alternatives.__opts__, {"test": True}):
            assert alternatives.set_(name, path) == ret

        comt = f"Alternative for {name} not updated"
        ret.update({"comment": comt, "result": True})
        with patch.dict(alternatives.__opts__, {"test": False}):
            assert alternatives.set_(name, path) == ret

        comt = f"Alternative {path} for {name} doesn't exist"
        ret.update({"comment": comt, "result": False})
        assert alternatives.set_(name, path) == ret



import pytest
from pytestshellutils.utils.processes import terminate_process

import salt.utils.event
import salt.utils.stringutils


@pytest.mark.slow_test
def test_event_return(master_opts):
    evt = None
    try:
        evt = salt.utils.event.EventReturn(master_opts)
        evt.start()
    except TypeError as exc:
        if "object" in str(exc):
            pytest.fail(f"'{exc}' TypeError should have not been raised")
    finally:
        if evt is not None:
            terminate_process(evt.pid, kill_children=True)


def test_filter_cluster_peer():
    assert (
        salt.utils.event.EventReturn._filter(
            {"__peer_id": "foo", "tag": "salt/test", "data": {"foo": "bar"}},
        )
        is False
    )


def test_filter_no_allow_or_deny():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
        )
        is True
    )


def test_filter_not_allowed():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
            allow=["foo/*"],
        )
        is False
    )


def test_filter_not_denied():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
            deny=["foo/*"],
        )
        is True
    )


def test_filter_allowed():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
            allow=["salt/*"],
        )
        is True
    )


def test_filter_denied():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
            deny=["salt/*"],
        )
        is False
    )


def test_filter_allowed_but_denied():
    assert (
        salt.utils.event.EventReturn._filter(
            {"tag": "salt/test", "data": {"foo": "bar"}},
            allow=["salt/*"],
            deny=["salt/test"],
        )
        is False
    )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.powerpath as powerpath
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {powerpath: {}}


def test_license_present():
    """
    Test to ensures that the specified PowerPath license key is present
    on the host.
    """
    name = "mylic"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock_t = MagicMock(
        side_effect=[
            {"result": True, "output": name},
            {"result": False, "output": name},
        ]
    )
    mock = MagicMock(side_effect=[False, True, True, True, True])
    mock_l = MagicMock(return_value=[{"key": name}])
    with patch.dict(
        powerpath.__salt__,
        {
            "powerpath.has_powerpath": mock,
            "powerpath.list_licenses": mock_l,
            "powerpath.add_license": mock_t,
        },
    ):
        comt = "PowerPath is not installed."
        ret.update({"comment": comt, "result": False})
        assert powerpath.license_present(name) == ret

        comt = f"License key {name} already present"
        ret.update({"comment": comt, "result": True})
        assert powerpath.license_present(name) == ret

        with patch.dict(powerpath.__opts__, {"test": True}):
            comt = "License key Mylic is set to be added"
            ret.update({"comment": comt, "result": None, "name": "Mylic"})
            assert powerpath.license_present("Mylic") == ret

        with patch.dict(powerpath.__opts__, {"test": False}):
            ret.update({"comment": name, "result": True, "changes": {"Mylic": "added"}})
            assert powerpath.license_present("Mylic") == ret

            ret.update({"result": False, "changes": {}})
            assert powerpath.license_present("Mylic") == ret


def test_license_absent():
    """
    Test to ensures that the specified PowerPath license key is absent
    on the host.
    """
    name = "mylic"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock_t = MagicMock(
        side_effect=[
            {"result": True, "output": name},
            {"result": False, "output": name},
        ]
    )
    mock = MagicMock(side_effect=[False, True, True, True, True])
    mock_l = MagicMock(return_value=[{"key": "salt"}])
    with patch.dict(
        powerpath.__salt__,
        {
            "powerpath.has_powerpath": mock,
            "powerpath.list_licenses": mock_l,
            "powerpath.remove_license": mock_t,
        },
    ):
        comt = "PowerPath is not installed."
        ret.update({"comment": comt, "result": False})
        assert powerpath.license_absent(name) == ret

        comt = f"License key {name} not present"
        ret.update({"comment": comt, "result": True})
        assert powerpath.license_absent(name) == ret

        with patch.dict(powerpath.__opts__, {"test": True}):
            comt = "License key salt is set to be removed"
            ret.update({"comment": comt, "result": None, "name": "salt"})
            assert powerpath.license_absent("salt") == ret

        with patch.dict(powerpath.__opts__, {"test": False}):
            ret.update(
                {"comment": name, "result": True, "changes": {"salt": "removed"}}
            )
            assert powerpath.license_absent("salt") == ret

            ret.update({"result": False, "changes": {}})
            assert powerpath.license_absent("salt") == ret



"""
Generic REST API SDB Module

:maintainer:    SaltStack
:maturity:      New
:platform:      all

.. versionadded:: 2015.8.0

This module allows access to a REST interface using an ``sdb://`` URI.

Like all REST modules, the REST module requires a configuration profile to
be configured in either the minion or master configuration file. This profile
requires very little. In the example:

.. code-block:: yaml

    my-rest-api:
      driver: rest
      urls:
        url: https://api.github.com/
      keys:
        url: https://api.github.com/users/{{user}}/keys
        backend: requests

The ``driver`` refers to the REST module, and must be set to ``rest`` in order
to use this driver. Each of the other items inside this block refers to a
separate set of HTTP items, including a URL and any options associated with it.
The options used here should match the options available in
``salt.utils.http.query()``.

In order to call the ``urls`` item in the example, the following reference can
be made inside a configuration file:

.. code-block:: yaml

    github_urls: sdb://my-rest-api/urls

Key/Value pairs may also be used with this driver, and merged into the URL using
the configured renderer (``jinja``, by default). For instance, in order to use
the ``keys`` item in the example, the following reference can be made:

.. code-block:: yaml

    github_urls: sdb://my-rest-api/keys?user=myuser

This will cause the following URL to actually be called:

.. code-block:: yaml

    https://api.github.com/users/myuser/keys

Key/Value pairs will NOT be passed through as GET data. If GET data needs to be
sent to the URL, then it should be configured in the SDB configuration block.
For instance:

.. code-block:: yaml

    another-rest-api:
      driver: rest
      user_data:
        url: https://api.example.com/users/
        params:
          user: myuser
"""

import logging

import salt.loader
import salt.utils.http as http
from salt.template import compile_template

log = logging.getLogger(__name__)

__func_alias__ = {"set_": "set"}


def set_(key, value, service=None, profile=None):  # pylint: disable=W0613
    """
    Set a key/value pair in the REST interface
    """
    return query(key, value, service, profile)


def get(key, service=None, profile=None):  # pylint: disable=W0613
    """
    Get a value from the REST interface
    """
    return query(key, None, service, profile)


def query(key, value=None, service=None, profile=None):  # pylint: disable=W0613
    """
    Get a value from the REST interface
    """
    comps = key.split("?")
    key = comps[0]
    key_vars = {}
    for pair in comps[1].split("&"):
        pair_key, pair_val = pair.split("=")
        key_vars[pair_key] = pair_val

    renderer = __opts__.get("renderer", "jinja|yaml")
    rend = salt.loader.render(__opts__, {})
    blacklist = __opts__.get("renderer_blacklist")
    whitelist = __opts__.get("renderer_whitelist")
    url = compile_template(
        ":string:",
        rend,
        renderer,
        blacklist,
        whitelist,
        input_data=profile[key]["url"],
        **key_vars
    )

    extras = {}
    for item in profile[key]:
        if item not in ("backend", "url"):
            extras[item] = profile[key][item]

    result = http.query(url, decode=True, **extras)

    return result["dict"]



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


def test_state_with_import(salt_ssh_cli, state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run("state.sls", "test")
    assert ret.returncode == 0
    assert ret.data


def test_state_with_import_from_dir(salt_ssh_cli, nested_state_tree):
    """
    verify salt-ssh can use imported map files in states
    """
    ret = salt_ssh_cli.run(
        "--extra-filerefs=salt://foo/map.jinja", "state.apply", "foo"
    )
    assert ret.returncode == 0
    assert ret.data


def test_state_low(salt_ssh_cli):
    """
    test state.low with salt-ssh
    """
    ret = salt_ssh_cli.run(
        "state.low", '{"state": "cmd", "fun": "run", "name": "echo blah"}'
    )
    assert ret.data["cmd_|-echo blah_|-echo blah_|-run"]["changes"]["stdout"] == "blah"


def test_state_high(salt_ssh_cli):
    """
    test state.high with salt-ssh
    """
    ret = salt_ssh_cli.run("state.high", '{"echo blah": {"cmd": ["run"]}}')
    assert ret.data["cmd_|-echo blah_|-echo blah_|-run"]["changes"]["stdout"] == "blah"


def test_state_test(salt_ssh_cli, state_tree):
    ret = salt_ssh_cli.run("state.test", "test")
    assert ret.returncode == 0
    assert ret.data
    assert (
        ret.data["test_|-Ok with def_|-Ok with def_|-succeed_with_changes"]["result"]
        is None
    )



"""
Minion side functions for salt-cp
"""

import base64
import errno
import fnmatch
import logging
import os
import urllib.parse

import salt.channel.client
import salt.crypt
import salt.fileclient
import salt.minion
import salt.utils.data
import salt.utils.files
import salt.utils.functools
import salt.utils.gzip_util
import salt.utils.path
import salt.utils.templates
import salt.utils.url
from salt.exceptions import CommandExecutionError
from salt.loader.dunder import (
    __context__,
    __file_client__,
    __grains__,
    __opts__,
    __pillar__,
)

log = logging.getLogger(__name__)

__proxyenabled__ = ["*"]


def _auth():
    """
    Return the auth object
    """
    if "auth" not in __context__:
        __context__["auth"] = salt.crypt.SAuth(__opts__)
    return __context__["auth"]


def _gather_pillar(pillarenv, pillar_override):
    """
    Whenever a state run starts, gather the pillar data fresh
    """
    pillar = salt.pillar.get_pillar(
        __opts__,
        __grains__.value(),
        __opts__["id"],
        __opts__["saltenv"],
        pillar_override=pillar_override,
        pillarenv=pillarenv,
    )
    ret = pillar.compile_pillar()
    if pillar_override and isinstance(pillar_override, dict):
        ret.update(pillar_override)
    return ret


def recv(files, dest):
    """
    Used with salt-cp, pass the files dict, and the destination.

    This function receives small fast copy files from the master via salt-cp.
    It does not work via the CLI.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.recv
    """
    ret = {}
    for path, data in files.items():
        if os.path.basename(path) == os.path.basename(dest) and not os.path.isdir(dest):
            final = dest
        elif os.path.isdir(dest):
            final = os.path.join(dest, os.path.basename(path))
        elif os.path.isdir(os.path.dirname(dest)):
            final = dest
        else:
            return "Destination unavailable"

        try:
            with salt.utils.files.fopen(final, "w+") as fp_:
                fp_.write(data)
            ret[final] = True
        except OSError:
            ret[final] = False

    return ret


def recv_chunked(dest, chunk, append=False, compressed=True, mode=None):
    """
    This function receives files copied to the minion using ``salt-cp`` and is
    not intended to be used directly on the CLI.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.recv_chunked
    """
    if "retcode" not in __context__:
        __context__["retcode"] = 0

    def _error(msg):
        __context__["retcode"] = 1
        return msg

    if chunk is None:
        # dest is an empty dir and needs to be created
        try:
            os.makedirs(dest)
        except OSError as exc:
            if exc.errno == errno.EEXIST:
                if os.path.isfile(dest):
                    return "Path exists and is a file"
            else:
                return _error(str(exc))
        return True

    chunk = base64.b64decode(chunk)

    open_mode = "ab" if append else "wb"
    try:
        fh_ = salt.utils.files.fopen(dest, open_mode)  # pylint: disable=W8470
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            # Parent dir does not exist, we need to create it
            return _error(str(exc))
        try:
            os.makedirs(os.path.dirname(dest))
        except OSError as makedirs_exc:
            # Failed to make directory
            return _error(str(makedirs_exc))
        fh_ = salt.utils.files.fopen(dest, open_mode)  # pylint: disable=W8470

    try:
        # Write the chunk to disk
        fh_.write(salt.utils.gzip_util.uncompress(chunk) if compressed else chunk)
    except OSError as exc:
        # Write failed
        return _error(str(exc))
    else:
        # Write successful
        if not append and mode is not None:
            # If this is the first chunk we're writing, set the mode
            # log.debug('Setting mode for %s to %s', dest, oct(mode))
            log.debug("Setting mode for %s to %s", dest, mode)
            try:
                os.chmod(dest, mode)
            except OSError as exc:
                return _error(str(exc))
        return True
    finally:
        try:
            fh_.close()
        except AttributeError:
            pass


def _client():
    """
    Return a file client

    If the __file_client__ context is set return it, otherwize create a new
    file client using __opts__.
    """
    if __file_client__:
        return __file_client__.value()
    return salt.fileclient.get_file_client(__opts__.value())


def _render_filenames(path, dest, saltenv, template, **kw):
    """
    Process markup in the :param:`path` and :param:`dest` variables (NOT the
    files under the paths they ultimately point to) according to the markup
    format provided by :param:`template`.
    """
    if not template:
        return (path, dest)

    # render the path as a template using path_template_engine as the engine
    if template not in salt.utils.templates.TEMPLATE_REGISTRY:
        raise CommandExecutionError(
            f"Attempted to render file paths with unavailable engine {template}"
        )

    kwargs = {}
    kwargs["salt"] = __salt__
    if "pillarenv" in kw or "pillar" in kw:
        pillarenv = kw.get("pillarenv", __opts__.get("pillarenv"))
        kwargs["pillar"] = _gather_pillar(pillarenv, kw.get("pillar"))
    else:
        kwargs["pillar"] = __pillar__
    kwargs["grains"] = __grains__
    kwargs["opts"] = __opts__
    kwargs["saltenv"] = saltenv

    def _render(contents):
        """
        Render :param:`contents` into a literal pathname by writing it to a
        temp file, rendering that file, and returning the result.
        """
        # write out path to temp file
        tmp_path_fn = salt.utils.files.mkstemp()
        with salt.utils.files.fopen(tmp_path_fn, "w+") as fp_:
            fp_.write(salt.utils.stringutils.to_str(contents))
        data = salt.utils.templates.TEMPLATE_REGISTRY[template](
            tmp_path_fn, to_str=True, **kwargs
        )
        salt.utils.files.safe_rm(tmp_path_fn)
        if not data["result"]:
            # Failed to render the template
            raise CommandExecutionError(
                "Failed to render file path with error: {}".format(data["data"])
            )
        else:
            return data["data"]

    path = _render(path)
    dest = _render(dest)
    return (path, dest)


def get_file(
    path, dest, saltenv=None, makedirs=False, template=None, gzip=None, **kwargs
):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    .. versionchanged:: 2018.3.0
        ``dest`` can now be a directory

    Used to get a single file from the salt master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_file salt://path/to/file /minion/dest

    Template rendering can be enabled on both the source and destination file
    names like so:

    .. code-block:: bash

        salt '*' cp.get_file "salt://{{grains.os}}/vimrc" /etc/vimrc template=jinja

    This example would instruct all Salt minions to download the vimrc from a
    directory with the same name as their os grain and copy it to /etc/vimrc

    For larger files, the cp.get_file module also supports gzip compression.
    Because gzip is CPU-intensive, this should only be used in scenarios where
    the compression ratio is very high (e.g. pretty-printed JSON or YAML
    files).

    Use the *gzip* named argument to enable it.  Valid values are 1..9, where 1
    is the lightest compression and 9 the heaviest.  1 uses the least CPU on
    the master (and minion), 9 uses the most.

    There are two ways of defining the fileserver environment (a.k.a.
    ``saltenv``) from which to retrieve the file. One is to use the ``saltenv``
    parameter, and the other is to use a querystring syntax in the ``salt://``
    URL. The below two examples are equivalent:

    .. code-block:: bash

        salt '*' cp.get_file salt://foo/bar.conf /etc/foo/bar.conf saltenv=config
        salt '*' cp.get_file salt://foo/bar.conf?saltenv=config /etc/foo/bar.conf

    .. note::
        It may be necessary to quote the URL when using the querystring method,
        depending on the shell being used to run the command.
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    (path, dest) = _render_filenames(path, dest, saltenv, template, **kwargs)

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    if not hash_file(path, saltenv):
        return ""
    else:
        with _client() as client:
            return client.get_file(path, dest, makedirs, saltenv, gzip)


def envs():
    """
    List available environments for fileserver

    CLI Example:

    .. code-block:: bash

        salt '*' cp.envs
    """
    with _client() as client:
        return client.envs()


def get_template(path, dest, template="jinja", saltenv=None, makedirs=False, **kwargs):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Render a file as a template before setting it down.
    Warning, order is not the same as in fileclient.cp for
    non breaking old API.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_template salt://path/to/template /minion/dest
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    if "salt" not in kwargs:
        kwargs["salt"] = __salt__
    if "pillar" not in kwargs:
        kwargs["pillar"] = __pillar__
    if "grains" not in kwargs:
        kwargs["grains"] = __grains__
    if "opts" not in kwargs:
        kwargs["opts"] = __opts__
    with _client() as client:
        return client.get_template(path, dest, template, makedirs, saltenv, **kwargs)


def get_dir(path, dest, saltenv=None, template=None, gzip=None, **kwargs):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Used to recursively copy a directory from the salt master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_dir salt://path/to/dir/ /minion/dest

    get_dir supports the same template and gzip arguments as get_file.
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    (path, dest) = _render_filenames(path, dest, saltenv, template, **kwargs)

    with _client() as client:
        return client.get_dir(path, dest, saltenv, gzip)


def get_url(path, dest="", saltenv=None, makedirs=False, source_hash=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    .. versionchanged:: 2018.3.0
        ``dest`` can now be a directory

    Used to get a single file from a URL.

    path
        A URL to download a file from. Supported URL schemes are: ``salt://``,
        ``http://``, ``https://``, ``ftp://``, ``s3://``, ``swift://`` and
        ``file://`` (local filesystem). If no scheme was specified, this is
        equivalent of using ``file://``.
        If a ``file://`` URL is given, the function just returns absolute path
        to that file on a local filesystem.
        The function returns ``False`` if Salt was unable to fetch a file from
        a ``salt://`` URL.

    dest
        The default behaviour is to write the fetched file to the given
        destination path. If this parameter is omitted or set as empty string
        (``''``), the function places the remote file on the local filesystem
        inside the Minion cache directory and returns the path to that file.

        .. note::

            To simply return the file contents instead, set destination to
            ``None``. This works with ``salt://``, ``http://``, ``https://``
            and ``file://`` URLs. The files fetched by ``http://`` and
            ``https://`` will not be cached.

    saltenv
        Salt fileserver environment from which to retrieve the file. Ignored if
        ``path`` is not a ``salt://`` URL.

    source_hash
        If ``path`` is an http(s) or ftp URL and the file exists in the
        minion's file cache, this option can be passed to keep the minion from
        re-downloading the file if the cached copy matches the specified hash.

        .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_url salt://my/file /tmp/this_file_is_mine
        salt '*' cp.get_url http://www.slashdot.org /tmp/index.html
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    if isinstance(dest, str):
        with _client() as client:
            result = client.get_url(
                path, dest, makedirs, saltenv, source_hash=source_hash
            )
    else:

        with _client() as client:
            result = client.get_url(
                path, None, makedirs, saltenv, no_cache=True, source_hash=source_hash
            )
    if not result:
        log.error(
            "Unable to fetch file %s from saltenv %s.",
            salt.utils.url.redact_http_basic_auth(path),
            saltenv,
        )
    if result:
        return salt.utils.stringutils.to_unicode(result)
    return result


def get_file_str(path, saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Download a file from a URL to the Minion cache directory and return the
    contents of that file

    Returns ``False`` if Salt was unable to cache a file from a URL.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_file_str salt://my/file
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    fn_ = cache_file(path, saltenv)
    if isinstance(fn_, str):
        try:
            with salt.utils.files.fopen(fn_, "r") as fp_:
                return salt.utils.stringutils.to_unicode(fp_.read())
        except OSError:
            return False
    return fn_


def cache_file(path, saltenv=None, source_hash=None, verify_ssl=True, use_etag=False):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Used to cache a single file on the Minion

    Returns the location of the new cached file on the Minion

    source_hash
        If ``name`` is an http(s) or ftp URL and the file exists in the
        minion's file cache, this option can be passed to keep the minion from
        re-downloading the file if the cached copy matches the specified hash.

        .. versionadded:: 2018.3.0

    verify_ssl
        If ``False``, remote https file sources (``https://``) and source_hash
        will not attempt to validate the servers certificate. Default is True.

        .. versionadded:: 3002

    use_etag
        If ``True``, remote http/https file sources will attempt to use the
        ETag header to determine if the remote file needs to be downloaded.
        This provides a lightweight mechanism for promptly refreshing files
        changed on a web server without requiring a full hash comparison via
        the ``source_hash`` parameter.

        .. versionadded:: 3005

    CLI Example:

    .. code-block:: bash

        salt '*' cp.cache_file salt://path/to/file

    There are two ways of defining the fileserver environment (a.k.a.
    ``saltenv``) from which to cache the file. One is to use the ``saltenv``
    parameter, and the other is to use a querystring syntax in the ``salt://``
    URL. The below two examples are equivalent:

    .. code-block:: bash

        salt '*' cp.cache_file salt://foo/bar.conf saltenv=config
        salt '*' cp.cache_file salt://foo/bar.conf?saltenv=config

    If the path being cached is a ``salt://`` URI, and the path does not exist,
    then ``False`` will be returned.

    .. note::
        It may be necessary to quote the URL when using the querystring method,
        depending on the shell being used to run the command.
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    path = salt.utils.data.decode(path)
    saltenv = salt.utils.data.decode(saltenv)

    contextkey = "{}_|-{}_|-{}".format("cp.cache_file", path, saltenv)

    path_is_remote = (
        urllib.parse.urlparse(path).scheme in salt.utils.files.REMOTE_PROTOS
    )
    try:
        if path_is_remote and contextkey in __context__:
            # Prevent multiple caches in the same salt run. Affects remote URLs
            # since the master won't know their hash, so the fileclient
            # wouldn't be able to prevent multiple caches if we try to cache
            # the remote URL more than once.
            if os.path.isfile(__context__[contextkey]):
                return __context__[contextkey]
            else:
                # File is in __context__ but no longer exists in the minion
                # cache, get rid of the context key and re-cache below.
                # Accounts for corner case where file is removed from minion
                # cache between cp.cache_file calls in the same salt-run.
                __context__.pop(contextkey)
    except AttributeError:
        pass

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    with _client() as client:
        result = client.cache_file(
            path,
            saltenv,
            source_hash=source_hash,
            verify_ssl=verify_ssl,
            use_etag=use_etag,
        )
    if not result and not use_etag:
        log.error("Unable to cache file '%s' from saltenv '%s'.", path, saltenv)
    if path_is_remote:
        # Cache was successful, store the result in __context__ to prevent
        # multiple caches (see above).
        __context__[contextkey] = result
    return result


cache_file_ssh = salt.utils.functools.alias_function(cache_file, "cache_file_ssh")


def cache_dest(url, saltenv=None):
    """
    .. versionadded:: 3000

    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Returns the expected cache path for the file, if cached using
    :py:func:`cp.cache_file <salt.modules.cp.cache_file>`.

    .. note::
        This only returns the _expected_ path, it does not tell you if the URL
        is really cached. To check if the URL is cached, use
        :py:func:`cp.is_cached <salt.modules.cp.is_cached>` instead.

    CLI Examples:

    .. code-block:: bash

        salt '*' cp.cache_dest https://foo.com/bar.rpm
        salt '*' cp.cache_dest salt://my/file
        salt '*' cp.cache_dest salt://my/file saltenv=dev
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.cache_dest(url, saltenv)


def cache_files(paths, saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Used to gather many files from the Master, the gathered files will be
    saved in the minion cachedir reflective to the paths retrieved from the
    Master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.cache_files salt://pathto/file1,salt://pathto/file1

    There are two ways of defining the fileserver environment (a.k.a.
    ``saltenv``) from which to cache the files. One is to use the ``saltenv``
    parameter, and the other is to use a querystring syntax in the ``salt://``
    URL. The below two examples are equivalent:

    .. code-block:: bash

        salt '*' cp.cache_files salt://foo/bar.conf,salt://foo/baz.conf saltenv=config
        salt '*' cp.cache_files salt://foo/bar.conf?saltenv=config,salt://foo/baz.conf?saltenv=config

    The querystring method is less useful when all files are being cached from
    the same environment, but is a good way of caching files from multiple
    different environments in the same command. For example, the below command
    will cache the first file from the ``config1`` environment, and the second
    one from the ``config2`` environment.

    .. code-block:: bash

        salt '*' cp.cache_files salt://foo/bar.conf?saltenv=config1,salt://foo/bar.conf?saltenv=config2

    .. note::
        It may be necessary to quote the URL when using the querystring method,
        depending on the shell being used to run the command.
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.cache_files(paths, saltenv)


def cache_dir(
    path, saltenv=None, include_empty=False, include_pat=None, exclude_pat=None
):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Download and cache everything under a directory from the master


    include_pat : None
        Glob or regex to narrow down the files cached from the given path. If
        matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. versionadded:: 2014.7.0

    exclude_pat : None
        Glob or regex to exclude certain files from being cached from the given
        path. If matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. note::

            If used with ``include_pat``, files matching this pattern will be
            excluded from the subset of files defined by ``include_pat``.

        .. versionadded:: 2014.7.0

    CLI Examples:

    .. code-block:: bash

        salt '*' cp.cache_dir salt://path/to/dir
        salt '*' cp.cache_dir salt://path/to/dir include_pat='E@*.py$'
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.cache_dir(path, saltenv, include_empty, include_pat, exclude_pat)


def cache_master(saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Retrieve all of the files on the master and cache them locally

    CLI Example:

    .. code-block:: bash

        salt '*' cp.cache_master
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.cache_master(saltenv)


def cache_local_file(path):
    """
    Cache a local file on the minion in the localfiles cache

    CLI Example:

    .. code-block:: bash

        salt '*' cp.cache_local_file /etc/hosts
    """
    if not os.path.exists(path):
        return ""

    path_cached = is_cached(path)

    # If the file has already been cached, return the path
    if path_cached:
        path_hash = hash_file(path)
        path_cached_hash = hash_file(path_cached)

        if path_hash["hsum"] == path_cached_hash["hsum"]:
            return path_cached

    # The file hasn't been cached or has changed; cache it
    with _client() as client:
        return client.cache_local_file(path)


def list_states(saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    List all of the available state files in an environment

    CLI Example:

    .. code-block:: bash

        salt '*' cp.list_states
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.list_states(saltenv)


def list_master(saltenv=None, prefix=""):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    List all of the files stored on the master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.list_master
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.file_list(saltenv, prefix)


def list_master_dirs(saltenv=None, prefix=""):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    List all of the directories stored on the master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.list_master_dirs
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.dir_list(saltenv, prefix)


def list_master_symlinks(saltenv=None, prefix=""):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    List all of the symlinks stored on the master

    CLI Example:

    .. code-block:: bash

        salt '*' cp.list_master_symlinks
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.symlink_list(saltenv, prefix)


def list_minion(saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    List all of the files cached on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' cp.list_minion
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"
    with _client() as client:
        return client.file_local_list(saltenv)


def is_cached(path, saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Returns the full path to a file if it is cached locally on the minion
    otherwise returns a blank string

    CLI Example:

    .. code-block:: bash

        salt '*' cp.is_cached salt://path/to/file
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    with _client() as client:
        return client.is_cached(path, saltenv)


def hash_file(path, saltenv=None):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Return the hash of a file, to get the hash of a file on the
    salt master file server prepend the path with salt://<file on server>
    otherwise, prepend the file with / for a local file.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.hash_file salt://path/to/file
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    with _client() as client:
        return client.hash_file(path, saltenv)


hash_file_ssh = salt.utils.functools.alias_function(hash_file, "hash_file_ssh")


def stat_file(path, saltenv=None, octal=True):
    """
    .. versionchanged:: 3005
        ``saltenv`` will use value from config if not explicitly set

    Return the permissions of a file, to get the permissions of a file on the
    salt master file server prepend the path with salt://<file on server>
    otherwise, prepend the file with / for a local file.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.stat_file salt://path/to/file
    """
    if not saltenv:
        saltenv = __opts__["saltenv"] or "base"

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    with _client() as client:
        stat = client.hash_and_stat_file(path, saltenv)[1]
    if stat is None:
        return stat
    return salt.utils.files.st_mode_to_octal(stat[0]) if octal is True else stat[0]


def push(path, keep_symlinks=False, upload_path=None, remove_source=False):
    """
    WARNING Files pushed to the master will have global read permissions..

    Push a file from the minion up to the master, the file will be saved to
    the salt master in the master's minion files cachedir
    (defaults to ``/var/cache/salt/master/minions/minion-id/files``)

    Since this feature allows a minion to push a file up to the master server
    it is disabled by default for security purposes. To enable, set
    ``file_recv`` to ``True`` in the master configuration file, and restart the
    master.

    keep_symlinks
        Keep the path value without resolving its canonical form

    upload_path
        Provide a different path inside the master's minion files cachedir

    remove_source
        Remove the source file on the minion

        .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' cp.push /etc/fstab
        salt '*' cp.push /etc/system-release keep_symlinks=True
        salt '*' cp.push /etc/fstab upload_path='/new/path/fstab'
        salt '*' cp.push /tmp/filename remove_source=True
    """
    log.debug("Trying to copy '%s' to master", path)
    if "../" in path or not os.path.isabs(path):
        log.debug("Path must be absolute, returning False")
        return False
    if not keep_symlinks:
        path = os.path.realpath(path)
    if not os.path.isfile(path):
        log.debug("Path failed os.path.isfile check, returning False")
        return False
    auth = _auth()

    if upload_path:
        if "../" in upload_path:
            log.debug("Path must be absolute, returning False")
            log.debug("Bad path: %s", upload_path)
            return False
        load_path = upload_path.lstrip(os.sep)
    else:
        load_path = path.lstrip(os.sep)
    # Normalize the path. This does not eliminate
    # the possibility that relative entries will still be present
    load_path_normal = os.path.normpath(load_path)

    # If this is Windows and a drive letter is present, remove it
    load_path_split_drive = os.path.splitdrive(load_path_normal)[1]

    # Finally, split the remaining path into a list for delivery to the master
    load_path_list = [_f for _f in load_path_split_drive.split(os.sep) if _f]

    load = {
        "cmd": "_file_recv",
        "id": __opts__["id"],
        "path": load_path_list,
        "size": os.path.getsize(path),
        "tok": auth.gen_token(b"salt"),
    }

    with salt.channel.client.ReqChannel.factory(__opts__) as channel:
        with salt.utils.files.fopen(path, "rb") as fp_:
            init_send = False
            while True:
                load["loc"] = fp_.tell()
                load["data"] = fp_.read(__opts__["file_buffer_size"])
                if not load["data"] and init_send:
                    if remove_source:
                        try:
                            salt.utils.files.rm_rf(path)
                            log.debug("Removing source file '%s'", path)
                        except OSError:
                            log.error("cp.push failed to remove file '%s'", path)
                            return False
                    return True
                ret = channel.send(load)
                if not ret:
                    log.error(
                        "cp.push Failed transfer failed. Ensure master has "
                        "'file_recv' set to 'True' and that the file "
                        "is not larger than the 'file_recv_size_max' "
                        "setting on the master."
                    )
                    return ret
                init_send = True


def push_dir(path, glob=None, upload_path=None):
    """
    Push a directory from the minion up to the master, the files will be saved
    to the salt master in the master's minion files cachedir (defaults to
    ``/var/cache/salt/master/minions/minion-id/files``).  It also has a glob
    for matching specific files using globbing.

    .. versionadded:: 2014.7.0

    Since this feature allows a minion to push files up to the master server it
    is disabled by default for security purposes. To enable, set ``file_recv``
    to ``True`` in the master configuration file, and restart the master.

    upload_path
        Provide a different path and directory name inside the master's minion
        files cachedir

    CLI Example:

    .. code-block:: bash

        salt '*' cp.push /usr/lib/mysql
        salt '*' cp.push /usr/lib/mysql upload_path='/newmysql/path'
        salt '*' cp.push_dir /etc/modprobe.d/ glob='*.conf'
    """
    if "../" in path or not os.path.isabs(path):
        return False
    tmpupload_path = upload_path
    path = os.path.realpath(path)
    if os.path.isfile(path):
        return push(path, upload_path=upload_path)
    else:
        filelist = []
        for root, _, files in salt.utils.path.os_walk(path):
            filelist += [os.path.join(root, tmpfile) for tmpfile in files]
        if glob is not None:
            filelist = [
                fi for fi in filelist if fnmatch.fnmatch(os.path.basename(fi), glob)
            ]
        if not filelist:
            return False
        for tmpfile in filelist:
            if upload_path and tmpfile.startswith(path):
                tmpupload_path = os.path.join(
                    os.path.sep,
                    upload_path.strip(os.path.sep),
                    tmpfile.replace(path, "").strip(os.path.sep),
                )
            ret = push(tmpfile, upload_path=tmpupload_path)
            if not ret:
                return ret
    return True



import copy
import textwrap

import pytest
from saltfactories.utils import random_string

import salt.config
from tests.support.mock import MagicMock, patch


@pytest.mark.slow_test
def test_post_master_init_metaproxy_called(io_loop):
    """
    Tests that when the _post_master_ini function is called, _metaproxy_call is also called.
    """

    mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
    mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)
    mock_jid_queue = [123]
    proxy_minion = salt.minion.ProxyMinion(
        mock_opts,
        jid_queue=copy.copy(mock_jid_queue),
        io_loop=io_loop,
    )
    mock_metaproxy_call = MagicMock()
    with patch(
        "salt.minion._metaproxy_call",
        return_value=mock_metaproxy_call,
        autospec=True,
    ):
        try:
            ret = proxy_minion._post_master_init("dummy_master")
            salt.minion._metaproxy_call.assert_called_once()
        finally:
            proxy_minion.destroy()


@pytest.mark.slow_test
async def test_handle_decoded_payload_metaproxy_called(io_loop):
    """
    Tests that when the _handle_decoded_payload function is called, _metaproxy_call is also called.
    """
    mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
    mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)

    mock_data = {"fun": "foo.bar", "jid": 123}
    mock_jid_queue = [123]
    proxy_minion = salt.minion.ProxyMinion(
        mock_opts,
        jid_queue=copy.copy(mock_jid_queue),
        io_loop=io_loop,
    )

    async def mock_metaproxy_call(*args, **kwargs):
        mock_metaproxy_call.calls += 1

    mock_metaproxy_call.calls = 0
    with patch(
        "salt.minion._metaproxy_call",
        return_value=mock_metaproxy_call,
        autospec=True,
    ):
        try:
            ret = await proxy_minion._handle_decoded_payload(mock_data)
            assert proxy_minion.jid_queue, mock_jid_queue
            assert mock_metaproxy_call.calls == 1
        finally:
            proxy_minion.destroy()


@pytest.mark.slow_test
async def test_handle_payload_metaproxy_called(io_loop):
    """
    Tests that when the _handle_payload function is called, _metaproxy_call is also called.
    """
    mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
    mock_opts.update(salt.config.DEFAULT_PROXY_MINION_OPTS)

    mock_data = {"fun": "foo.bar", "jid": 123}
    mock_jid_queue = [123]
    proxy_minion = salt.minion.ProxyMinion(
        mock_opts,
        jid_queue=copy.copy(mock_jid_queue),
        io_loop=io_loop,
    )

    async def mock_metaproxy_call(*args, **kwargs):
        mock_metaproxy_call.calls += 1

    mock_metaproxy_call.calls = 0
    with patch(
        "salt.minion._metaproxy_call",
        return_value=mock_metaproxy_call,
        autospec=True,
    ):
        try:
            ret = await proxy_minion._handle_decoded_payload(mock_data)
            assert proxy_minion.jid_queue == mock_jid_queue
            assert mock_metaproxy_call.calls == 1
        finally:
            proxy_minion.destroy()


def test_proxy_config_default_include(tmp_path):
    """
    Tests that when the proxy_config function is called,
    for the proxy minion, eg. /etc/salt/proxy.d/<The-Proxy-ID>/*.conf
    """
    proxyid = random_string("proxy-")
    root_dir = tmp_path / "root"
    conf_dir = root_dir / "conf"
    conf_file = conf_dir / "proxy"
    conf_d_dir = conf_dir / "proxy.d"
    proxy_conf_d = conf_d_dir / proxyid
    proxy_conf_d.mkdir(parents=True)

    with salt.utils.files.fopen(str(conf_file), "w") as wfh:
        wfh.write(
            textwrap.dedent(
                """\
                id: {id}
                root_dir: {root_dir}
                pidfile: run/proxy.pid
                pki_dir: pki
                cachedir: cache
                sock_dir: run/proxy
                log_file: logs/proxy.log
                """.format(
                    id=proxyid, root_dir=root_dir
                )
            )
        )

    with salt.utils.files.fopen(str(proxy_conf_d / "_schedule.conf"), "w") as wfh:
        wfh.write(
            textwrap.dedent(
                """\
                schedule:
                  test_job:
                    args: [arg1, arg2]
                    enabled: true
                    function: test.arg
                    jid_include: true
                    kwargs: {key1: value1, key2: value2}
                    maxrunning: 1
                    name: test_job
                    return_job: false
                """
            )
        )
    opts = salt.config.proxy_config(
        str(conf_file),
        minion_id=proxyid,
        cache_minion_id=False,
    )
    assert "schedule" in opts
    assert "test_job" in opts["schedule"]



"""
Unit tests for salt.transport.base.
"""

import ssl

import pytest

import salt.transport.base
from tests.support.mock import patch

pytestmark = [
    pytest.mark.core_test,
]


def test_unclosed_warning():

    transport = salt.transport.base.Transport()
    assert transport._closing is False
    assert transport._connect_called is False
    transport.connect()
    assert transport._connect_called is True
    with pytest.warns(salt.transport.base.TransportWarning):
        del transport


@patch("ssl.SSLContext")
def test_ssl_context_legacy_opts(mock):
    ctx = salt.transport.base.ssl_context(
        {
            "certfile": "server.crt",
            "keyfile": "server.key",
            "cert_reqs": "CERT_NONE",
            "ca_certs": "ca.crt",
        }
    )
    ctx.load_cert_chain.assert_called_with(
        "server.crt",
        "server.key",
    )
    ctx.load_verify_locations.assert_called_with("ca.crt")
    assert ssl.VerifyMode.CERT_NONE == ctx.verify_mode
    assert not ctx.check_hostname


@patch("ssl.SSLContext")
def test_ssl_context_opts(mock):
    mock.verify_flags = ssl.VerifyFlags.VERIFY_X509_TRUSTED_FIRST
    ctx = salt.transport.base.ssl_context(
        {
            "certfile": "server.crt",
            "keyfile": "server.key",
            "cert_reqs": "CERT_OPTIONAL",
            "verify_locations": [
                "ca.crt",
                {"cafile": "crl.pem"},
                {"capath": "/tmp/mycapathsdf"},
                {"cadata": "mycadataother"},
                {"CADATA": "mycadatasdf"},
            ],
            "verify_flags": [
                "VERIFY_CRL_CHECK_CHAIN",
            ],
        }
    )
    ctx.load_cert_chain.assert_called_with(
        "server.crt",
        "server.key",
    )
    ctx.load_verify_locations.assert_any_call(cafile="ca.crt")
    ctx.load_verify_locations.assert_any_call(cafile="crl.pem")
    ctx.load_verify_locations.assert_any_call(capath="/tmp/mycapathsdf")
    ctx.load_verify_locations.assert_any_call(cadata="mycadataother")
    ctx.load_verify_locations.assert_called_with(cadata="mycadatasdf")
    assert ssl.VerifyMode.CERT_OPTIONAL == ctx.verify_mode
    assert ctx.check_hostname
    assert ssl.VerifyFlags.VERIFY_CRL_CHECK_CHAIN & ctx.verify_flags



"""
Simple Smoke Tests for Connected Proxy Minion
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.mark.core_test
def test_can_it_ping(salt_cli, salt_proxy):
    """
    Ensure the proxy can ping
    """
    ret = salt_cli.run("test.ping", minion_tgt=salt_proxy.id)
    assert ret.data is True


@pytest.mark.slow_test
def test_list_pkgs(salt_cli, salt_proxy):
    """
    Package test 1, really just tests that the virtual function capability
    is working OK.
    """
    ret = salt_cli.run("pkg.list_pkgs", minion_tgt=salt_proxy.id)
    assert "coreutils" in ret.data
    assert "apache" in ret.data
    assert "redbull" in ret.data


def test_install_pkgs(salt_cli, salt_proxy):
    """
    Package test 2, really just tests that the virtual function capability
    is working OK.
    """
    ret = salt_cli.run("pkg.install", "thispkg", minion_tgt=salt_proxy.id)
    assert ret.data["thispkg"] == "1.0"

    ret = salt_cli.run("pkg.list_pkgs", minion_tgt=salt_proxy.id)

    assert ret.data["apache"] == "2.4"
    assert ret.data["redbull"] == "999.99"
    assert ret.data["thispkg"] == "1.0"


def test_remove_pkgs(salt_cli, salt_proxy):
    ret = salt_cli.run("pkg.remove", "apache", minion_tgt=salt_proxy.id)
    assert "apache" not in ret.data


def test_upgrade(salt_cli, salt_proxy):
    ret = salt_cli.run("pkg.upgrade", minion_tgt=salt_proxy.id)
    assert ret.data["coreutils"]["new"] == "2.0"
    assert ret.data["redbull"]["new"] == "1000.99"


def test_service_list(salt_cli, salt_proxy):
    ret = salt_cli.run("service.list", minion_tgt=salt_proxy.id)
    assert "ntp" in ret.data


def test_service_stop(salt_cli, salt_proxy):
    ret = salt_cli.run("service.stop", "ntp", minion_tgt=salt_proxy.id)
    ret = salt_cli.run("service.status", "ntp", minion_tgt=salt_proxy.id)
    assert ret.data is False


def test_service_start(salt_cli, salt_proxy):
    ret = salt_cli.run("service.start", "samba", minion_tgt=salt_proxy.id)
    ret = salt_cli.run("service.status", "samba", minion_tgt=salt_proxy.id)
    assert ret.data is True


def test_service_get_all(salt_cli, salt_proxy):
    ret = salt_cli.run("service.get_all", minion_tgt=salt_proxy.id)
    assert ret.data
    assert "samba" in ret.data


def test_grains_items(salt_cli, salt_proxy):
    ret = salt_cli.run("grains.items", minion_tgt=salt_proxy.id)
    assert ret.data["kernel"] == "proxy"
    assert ret.data["kernelrelease"] == "proxy"


def test_state_apply(salt_master, salt_cli, salt_proxy, tmp_path):
    test_file = tmp_path / "testfile"
    core_state = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        test_file
    )

    with salt_master.state_tree.base.temp_file("core.sls", core_state):
        ret = salt_cli.run("state.apply", "core", minion_tgt=salt_proxy.id)
        for value in ret.data.values():
            assert value["result"] is True


@pytest.mark.slow_test
def test_state_highstate(salt_master, salt_cli, salt_proxy, tmp_path):
    test_file = tmp_path / "testfile"
    top_sls = """
    base:
      '*':
        - core
        """

    core_state = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        test_file
    )

    with salt_master.state_tree.base.temp_file(
        "top.sls", top_sls
    ), salt_master.state_tree.base.temp_file("core.sls", core_state):
        ret = salt_cli.run("state.highstate", minion_tgt=salt_proxy.id)
        for value in ret.data.values():
            assert value["result"] is True



"""
Support for Debconf
"""

import logging
import os
import re

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
import salt.utils.versions

log = logging.getLogger(__name__)

__func_alias__ = {"set_": "set"}

# Define the module's virtual name
__virtualname__ = "debconf"


def __virtual__():
    """
    Confirm this module is on a Debian based system and that debconf-utils
    is installed.
    """
    if __grains__["os_family"] != "Debian":
        return (
            False,
            "The debconfmod module could not be loaded: unsupported OS family",
        )

    if salt.utils.path.which("debconf-get-selections") is None:
        return (
            False,
            "The debconfmod module could not be loaded: "
            "debconf-utils is not installed.",
        )

    return __virtualname__


def _unpack_lines(out):
    """
    Unpack the debconf lines
    """
    rexp = (
        "(?ms)"
        "^(?P<package>[^#]\\S+)[\t ]+"
        "(?P<question>\\S+)[\t ]+"
        "(?P<type>\\S+)[\t ]+"
        "(?P<value>[^\n]*)$"
    )
    lines = re.findall(rexp, out)
    return lines


def get_selections(fetchempty=True):
    """
    Answers to debconf questions for all packages in the following format::

        {'package': [['question', 'type', 'value'], ...]}

    CLI Example:

    .. code-block:: bash

        salt '*' debconf.get_selections
    """
    selections = {}
    cmd = "debconf-get-selections"

    out = __salt__["cmd.run_stdout"](cmd)

    lines = _unpack_lines(out)

    for line in lines:
        package, question, type_, value = line
        if fetchempty or value:
            (selections.setdefault(package, []).append([question, type_, value]))

    return selections


def show(name):
    """
    Answers to debconf questions for a package in the following format::

        [['question', 'type', 'value'], ...]

    If debconf doesn't know about a package, we return None.

    CLI Example:

    .. code-block:: bash

        salt '*' debconf.show <package name>
    """
    selections = get_selections()

    result = selections.get(name)
    return result


def _set_file(path):
    """
    Execute the set selections command for debconf
    """
    cmd = f"debconf-set-selections {path}"

    __salt__["cmd.run_stdout"](cmd, python_shell=False)


def set_(package, question, type, value, *extra):
    """
    Set answers to debconf questions for a package.

    CLI Example:

    .. code-block:: bash

        salt '*' debconf.set <package> <question> <type> <value> [<value> ...]
    """

    if extra:
        value = " ".join((value,) + tuple(extra))

    fd_, fname = salt.utils.files.mkstemp(prefix="salt-", close_fd=False)

    line = f"{package} {question} {type} {value}"
    os.write(fd_, salt.utils.stringutils.to_bytes(line))
    os.close(fd_)

    _set_file(fname)

    os.unlink(fname)

    return True


def set_template(path, template, context, defaults, saltenv="base", **kwargs):
    """
    Set answers to debconf questions from a template.

    path
        location of the file containing the package selections

    template
        template format

    context
        variables to add to the template environment

    default
        default values for the template environment

    CLI Example:

    .. code-block:: bash

        salt '*' debconf.set_template salt://pathto/pkg.selections.jinja jinja None None

    """

    path = __salt__["cp.get_template"](
        path=path,
        dest=None,
        template=template,
        saltenv=saltenv,
        context=context,
        defaults=defaults,
        **kwargs,
    )

    return set_file(path, saltenv, **kwargs)


def set_file(path, saltenv="base", **kwargs):
    """
    Set answers to debconf questions from a file.

    CLI Example:

    .. code-block:: bash

        salt '*' debconf.set_file salt://pathto/pkg.selections
    """
    if "__env__" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("__env__")

    path = __salt__["cp.cache_file"](path, saltenv)
    if path:
        _set_file(path)
        return True

    return False



import pytest

import salt.utils.files

from . import TEST_KEY


@pytest.fixture
def key_to_test(tmp_path):
    key_path = tmp_path / "cryptodom-3.4.6.pub"
    with salt.utils.files.fopen(key_path, "wb") as fd:
        fd.write(TEST_KEY.encode())
    return key_path



import pytest

from salt.runners import state as state_runner
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {state_runner: {"__opts__": {}, "__jid_event__": Mock()}}


def test_orchestrate_single_passes_pillar():
    """
    test state.orchestrate_single passes given pillar to state.single
    """
    mock_master_minion = Mock()
    mock_state_single = Mock()
    mock_master_minion.functions = {"state.single": mock_state_single}
    mock_master_minion.opts = {"id": "dummy"}
    test_pillar = {"test_entry": "exists"}
    with patch("salt.minion.MasterMinion", Mock(return_value=mock_master_minion)):
        state_runner.orchestrate_single(
            fun="pillar.get", name="test_entry", pillar=test_pillar
        )
        assert mock_state_single.call_args.kwargs["pillar"] == test_pillar


def test_orchestrate_single_does_not_pass_none_pillar():
    """
    test state.orchestrate_single does not pass pillar=None to state.single
    """
    mock_master_minion = Mock()
    mock_state_single = Mock()
    mock_master_minion.functions = {"state.single": mock_state_single}
    mock_master_minion.opts = {"id": "dummy"}
    with patch("salt.minion.MasterMinion", Mock(return_value=mock_master_minion)):
        state_runner.orchestrate_single(
            fun="pillar.get", name="test_entry", pillar=None
        )
        assert "pillar" not in mock_state_single.call_args.kwargs



"""
A module written originally by Armin Ronacher to manage file transfers in an
atomic way
"""

import errno
import os
import random
import shutil
import sys
import tempfile
import time

import salt.utils.files
import salt.utils.win_dacl

CAN_RENAME_OPEN_FILE = False
if os.name == "nt":  # pragma: no cover

    def _rename(src, dst):
        return False

    def _rename_atomic(src, dst):
        return False

    try:
        import ctypes

        _MOVEFILE_REPLACE_EXISTING = 0x1
        _MOVEFILE_WRITE_THROUGH = 0x8
        _MoveFileEx = ctypes.windll.kernel32.MoveFileExW  # pylint: disable=C0103

        def _rename(src, dst):  # pylint: disable=E0102
            if not isinstance(src, str):
                src = str(src, sys.getfilesystemencoding())
            if not isinstance(dst, str):
                dst = str(dst, sys.getfilesystemencoding())
            if _rename_atomic(src, dst):
                return True
            retry = 0
            rval = False
            while not rval and retry < 100:
                rval = _MoveFileEx(
                    src, dst, _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH
                )
                if not rval:
                    time.sleep(0.001)
                    retry += 1
            return rval

        # new in Vista and Windows Server 2008
        # pylint: disable=C0103
        _CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
        _CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
        _MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
        _CloseHandle = ctypes.windll.kernel32.CloseHandle
        # pylint: enable=C0103
        CAN_RENAME_OPEN_FILE = True

        def _rename_atomic(src, dst):  # pylint: disable=E0102
            tra = _CreateTransaction(None, 0, 0, 0, 0, 1000, "Atomic rename")
            if tra == -1:
                return False
            try:
                retry = 0
                rval = False
                while not rval and retry < 100:
                    rval = _MoveFileTransacted(
                        src,
                        dst,
                        None,
                        None,
                        _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH,
                        tra,
                    )
                    if rval:
                        rval = _CommitTransaction(tra)
                        break
                    else:
                        time.sleep(0.001)
                        retry += 1
                return rval
            finally:
                _CloseHandle(tra)

    except Exception:  # pylint: disable=broad-except
        pass

    def atomic_rename(src, dst):
        # Try atomic or pseudo-atomic rename
        if _rename(src, dst):
            return
        # Fall back to "move away and replace"
        try:
            os.rename(src, dst)
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise
            old = f"{dst}-{random.randint(0, sys.maxint):08x}"
            os.rename(dst, old)
            os.rename(src, dst)
            try:
                os.unlink(old)
            except Exception:  # pylint: disable=broad-except
                pass

else:
    atomic_rename = os.rename  # pylint: disable=C0103
    CAN_RENAME_OPEN_FILE = True


class _AtomicWFile:
    """
    Helper class for :func:`atomic_open`.
    """

    def __init__(self, fhanle, tmp_filename, filename):
        self._fh = fhanle
        self._tmp_filename = tmp_filename
        self._filename = filename

    def __getattr__(self, attr):
        return getattr(self._fh, attr)

    def __enter__(self):
        return self

    def close(self):
        if self._fh.closed:
            return
        self._fh.close()
        if salt.utils.win_dacl.HAS_WIN32:
            if os.path.isfile(self._filename):
                salt.utils.win_dacl.copy_security(
                    source=self._filename, target=self._tmp_filename
                )
        else:
            if os.path.isfile(self._filename):
                shutil.copymode(self._filename, self._tmp_filename)
                st = os.stat(self._filename)
                os.chown(self._tmp_filename, st.st_uid, st.st_gid)
            else:
                # chmod file to default mode based on umask
                os.chmod(self._tmp_filename, 0o666 & ~salt.utils.files.get_umask())
        atomic_rename(self._tmp_filename, self._filename)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.close()
        else:
            self._fh.close()
            try:
                os.remove(self._tmp_filename)
            except OSError:
                pass

    def __repr__(self):
        return "<{} {}{}, mode {}>".format(
            self.__class__.__name__,
            self._fh.closed and "closed " or "",
            self._filename,
            self._fh.mode,
        )


def atomic_open(filename, mode="w"):
    """
    Works like a regular `open()` but writes updates into a temporary
    file instead of the given file and moves it over when the file is
    closed.  The file returned behaves as if it was a regular Python
    """
    if mode in ("r", "rb", "r+", "rb+", "a", "ab"):
        raise TypeError("Read or append modes don't work with atomic_open")
    kwargs = {
        "prefix": ".___atomic_write",
        "dir": os.path.dirname(filename),
        "delete": False,
    }
    if "b" not in mode:
        kwargs["newline"] = ""
    ntf = tempfile.NamedTemporaryFile(mode, **kwargs)
    return _AtomicWFile(ntf, ntf.name, filename)
